"""
    Adapted for PyQt6 from https://github.com/yjg30737/pyqt-checkbox-list-widget
"""

from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtWidgets import QListWidgetItem
from PyQt6 import QtWidgets


class ToolTipListWidget(QtWidgets.QListWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Note that *args, **kwargs is necessary to be able to
        # be promoted from Qt Designer
        self.__initUi()

    def __initUi(self):
        self.setMouseTracking(True)
        self.itemEntered.connect(self.__showToolTip)

    def addItem(self, aitem) -> None:
        item = ""
        text = ""
        if isinstance(aitem, str):
            item = QListWidgetItem()
            text = aitem
            item.setText(text)
        elif isinstance(aitem, QListWidgetItem):
            item = aitem
            text = item.text()
        self.setItemWidget(item, QtWidgets.QWidget())
        super().addItem(item)

    def __showToolTip(self, item: QListWidgetItem):
        text = item.text()
        text_width = self.fontMetrics().boundingRect(text).width()
        width = self.width()
        if text_width > width:
            item.setToolTip(text)
        else:
            item.setToolTip("")


class CheckBoxListWidget(ToolTipListWidget):
    checkedSignal = pyqtSignal(int, Qt.CheckState)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.itemChanged.connect(self.__sendCheckedSignal)

    def __sendCheckedSignal(self, item):
        r_idx = self.row(item)
        state = item.checkState()
        self.checkedSignal.emit(r_idx, state)

    def addItems(self, items) -> None:
        for item in items:
            self.addItem(item)

    def addItem(self, item) -> None:
        if isinstance(item, str):
            item = QListWidgetItem(item)
        item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
        item.setCheckState(Qt.CheckState.Unchecked)
        super().addItem(item)

    def toggleState(self, state):
        for i in range(self.count()):
            item = self.item(i)
            if item.checkState() != state:
                item.setCheckState(state)

    def getCheckedRows(self):
        return self.__getFlagRows(Qt.CheckState.Checked)

    def getUncheckedRows(self):
        return self.__getFlagRows(Qt.CheckState.Unchecked)

    def __getFlagRows(self, flag: Qt.CheckState):
        flag_lst = []
        for i in range(self.count()):
            item = self.item(i)
            if item.checkState() == flag:
                flag_lst.append(i)

        return flag_lst

    def removeCheckedRows(self):
        self.__removeFlagRows(Qt.CheckState.Checked)

    def removeUncheckedRows(self):
        self.__removeFlagRows(Qt.CheckState.Unchecked)

    def __removeFlagRows(self, flag):
        flag_lst = self.__getFlagRows(flag)
        flag_lst = reversed(flag_lst)
        for i in flag_lst:
            self.takeItem(i)



from PyQt6 import uic
from PyQt6 import QtWidgets
from PyQt6 import QtCore
from pathlib import Path
import logging
from zhunter import DIRS


log = logging.getLogger(__name__)


def select_file(parent, fname, file_type):
    if fname is not None and Path(fname).exists():
        line_dir = Path(fname).parent
    else:
        line_dir = QtCore.QDir.currentPath()
    dialog = QtWidgets.QFileDialog(parent)
    dialog.setWindowTitle("Open file")
    dialog.setNameFilter(file_type)
    dialog.setDirectory(str(line_dir))
    dialog.setFileMode(QtWidgets.QFileDialog.FileMode.ExistingFile)
    filename = None
    if dialog.exec() == QtWidgets.QDialog.DialogCode.Accepted:
        filename = dialog.selectedFiles()
    if filename:
        return str(filename[0])


class SelectLineListsDialog(QtWidgets.QDialog):
    def __init__(self, parent):
        super(SelectLineListsDialog, self).__init__(parent)
        self.setWindowTitle("Select line lists files")
        self.parent = parent
        uic.loadUi(DIRS["UI"] / "line_list_selection_dialog.ui", self)

        self.emission_textbox.setText(str(parent.fnames["emission_lines"]))
        self.absorption_textbox.setText(str(parent.fnames["absorption_lines"]))
        self.fname_em = parent.fnames["emission_lines"]
        self.fname_abs = parent.fnames["absorption_lines"]

        self.em_line_file_select_button.clicked.connect(self.select_emission)
        self.abs_line_file_select_button.clicked.connect(self.select_absorption)

        if self.exec() == QtWidgets.QDialog.DialogCode.Accepted:
            log.info("Updated line lists.")
            self.parent.fnames["emission_lines"] = self.fname_em
            self.parent.fnames["absorption_lines"] = self.fname_abs
            self.parent.load_line_lists(calc_ratio=False)

    def select_emission(self):
        fname = select_file(
            self, self.fname_em, file_type="(*.csv *.txt *.dat *.ecsv *gz)"
        )
        if fname:
            self.fname_em = Path(fname)
            self.emission_textbox.setText(str(self.fname_em))

    def select_absorption(self):
        fname = select_file(
            self, self.fname_abs, file_type="(*.csv *.txt *.dat *.ecsv *gz)"
        )
        if fname:
            self.fname_abs = Path(fname)
            self.absorption_textbox.setText(str(self.fname_abs))



import pyqtgraph as pg

from PyQt6 import uic
from PyQt6 import QtWidgets
from PyQt6 import QtCore
from PyQt6.QtCore import Qt

from astropy.io.ascii import read as ascii_read
import astropy.units as u
from astropy.units.quantity import Quantity
import numpy as np

from pathlib import Path
from itertools import product
import logging
from zhunter import DIRS
from .line_list_selection import select_file

# from .CheckBoxListWidget import CheckBoxListWidget


log = logging.getLogger(__name__)


class VelocityPlot(QtWidgets.QMainWindow):
    def __init__(self, parent, z=None, lines=None):
        super(VelocityPlot, self).__init__(parent)
        uic.loadUi(DIRS["UI"] / "velocity_plot.ui", self)

        self.parent = parent

        # Colors
        self.colors = parent.colors
        # DONT FORGET TO REMOVE THIS
        # ITS A TEST TO SEE IF IT MATTERS OR IF ITS TOO LATE
        # pg.setConfigOption("foreground", self.colors["background"])
        # pg.setConfigOption("background", self.colors["foreground"])

        if lines is None:
            self.linelist_file_textbox.setText(str(parent.fnames["absorption_lines"]))
            self.fname = parent.fnames["absorption_lines"]
            self.load_lines()
        else:
            self.fname = None
            self.lines = lines

        self.linelistView.addItems(self.lines["name"])
        self.lines_to_plot = {}

        # Plot bounds
        self.wvlg_min = self.parent.data["wvlg_min"]
        self.wvlg_max = self.parent.data["wvlg_max"]

        # Connect signals and slots
        self.linelist_file_select_button.clicked.connect(self.select_linelist_file)
        self.plot_velocities_button.clicked.connect(self.update_plot)
        self.reset_button.clicked.connect(self.reset_plot_and_checklist)

        self.redshift = z

        if self.redshift is not None:
            self.textbox_for_z.setText(f"{self.redshift:.5f}")

        self.velLayout.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

    def select_linelist_file(self):
        fname = select_file(
            self, self.fname, file_type="(*.csv *.txt *.dat *.ecsv *gz)"
        )
        if fname:
            self.fname = Path(fname)
            self.linelist_file_textbox.setText(str(self.fname))
            self.load_lines()
            self.linelistView.clear()
            self.linelistView.addItems(self.lines["name"])

    def load_lines(self):
        try:
            self.lines = ascii_read(self.fname)
        except Exception as e:
            QtWidgets.QMessageBox.information(self, "Invalid input file", str(e))

    def reset_plot_and_checklist(self):
        self.velLayout.clear()
        self.linelistView.toggleState(Qt.CheckState.Unchecked)

    def update_plot(self):
        self.update_lines_to_plot()
        self.plot_velocities()

    def update_lines_to_plot(self):
        lnames = []
        for row in self.linelistView.getCheckedRows():
            lnames.append(self.linelistView.item(row).text())

        waves = (
            np.array([line["wave"] for line in self.lines if line["name"] in lnames])
            * self.lines["wave"].unit
        )
        waves_obs = waves * (1 + self.redshift)

        self.lines_to_plot = {
            n: {"name": n, "rest_wave": rw, "obs_wave": ow}
            for n, rw, ow in zip(lnames, waves, waves_obs)
            if (ow >= self.wvlg_min) and (ow <= self.wvlg_max)
        }

        if any([(ow < self.wvlg_min) or (ow > self.wvlg_max) for ow in waves_obs]):
            out_of_bounds = [
                n for n, ow in zip(lnames, waves_obs)
                if (ow < self.wvlg_min) and (ow > self.wvlg_max)
            ]
            QtWidgets.QMessageBox.information(
                self,
                "Lines out of bounds",
                f"Warning: could not plot lines {out_of_bounds} "
                "because they are outside of the bounds of the spectrum",
            )

    def plot_velocities(self):
        self.velLayout.clear()

        if not self.lines_to_plot:
            QtWidgets.QMessageBox.information(
                self,
                "No lines selected",
                "No lines are selected to be plotted in velocity space."
                " Please check at least one to show the plot.",
            )
            return

        wvlg = self.parent.data["wvlg_1D_disp"]
        flux = self.parent.data["flux_1D_disp"]
        unc = self.parent.data["unc_1D_disp"]

        n_rows, n_cols = self.__get_n_rows_and_cols()
        # take the first element of the list of line names
        # as the reference. All velocity plots will be linked
        # to this one
        ref_lname = list(self.lines_to_plot.keys())[0]

        vel_min = -600 * u.Unit("km/s")
        vel_max = 600 * u.Unit("km/s")

        # Create the plot
        for lines, indexes in zip(
            self.lines_to_plot.items(), product(range(1, n_cols + 1), range(n_rows))
        ):
            i, j = indexes
            lname, line = lines
            vel = wave_to_vel(wvlg, line["obs_wave"])
            imin = vel.searchsorted(vel_min)
            imax = vel.searchsorted(vel_max)

            vel_1D_spec = pg.PlotCurveItem(
                vel.value,
                flux.value,
                pen=pg.mkPen(color=self.colors["spec"]),
                stepMode="center",
            )
            vel_1D_unc = pg.PlotCurveItem(
                vel.value,
                unc.value,
                pen=pg.mkPen(color=self.colors["unc"]),
                stepMode="center",
            )

            line["pi"] = self.velLayout.addPlot(
                title=lname.replace("_", " "), name=lname, col=i, row=j
            )

            line["pi"].addItem(vel_1D_spec)
            line["pi"].addItem(vel_1D_unc)
            line["pi"].showGrid(x=True, y=True)
            line["pi"].hideButtons()
            line["pi"].setXLink(ref_lname)
            line["pi"].setXRange(vel_min.value, vel_max.value)
            try:
                line["pi"].setYRange(
                    np.min(flux[imin:imax].value),
                    np.max(flux[imin:imax].value),
                )
            except IndexError:
                pass
            line["pi"].vb.setLimits(xMin=np.min(vel.value), xMax=np.max(vel.value))

        self.velLayout.addLabel(
            "Velocity (km/s)", col=1, row=n_rows + 1, colspan=n_cols
        )
        self.velLayout.addLabel(
            "Flux" + f" ({flux.unit})", angle=-90, col=0, rowspan=n_rows
        )

        self.velLayout.setFocus()

    def __get_n_rows_and_cols(self):
        n_l = len(self.lines_to_plot.keys())
        if n_l <= 4:
            n_cols = 1
            n_rows = n_l
        else:
            n_cols = int(n_l / 2)
            n_rows = int(n_l / 2)
            if n_l % 2 != 0:
                n_rows += 1
        return n_rows, n_cols


class LineListModel(QtCore.QAbstractListModel):
    def __init__(self, *args, lines=None, **kwargs):
        super(LineListModel, self).__init__(*args, **kwargs)
        self.lines = lines or []

    def data(self, index, role):
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            # See below for the data structure.
            status, line = self.lines[index.row()]
            return f"{line['name']}"

    def rowCount(self, index):
        return len(self.lines)


def vel_to_wave(vel, w0):
    """Summary

    Parameters
    ----------
    vel : TYPE
        Description
    w0 : TYPE
        Description

    Returns
    -------
    TYPE
        Description
    """
    if isinstance(vel, Quantity):
        vel = vel.to(u.km / u.s)
    else:
        vel = vel * u.km / u.s
        log.info(
            "Input velocity is not a quantity so it has no units." " Assuming km/s."
        )

    if isinstance(w0, Quantity):
        w0 = w0.to(u.AA)
    else:
        log.info(
            "Input center wavelength is not a quantity so it has no units."
            " Assuming Angstrom."
        )
        w0 = w0 * u.AA

    wave = vel.to(u.AA, equivalencies=u.doppler_optical(w0))

    return wave if isinstance(vel, Quantity) else wave.value


def wave_to_vel(wave, w0):
    """Summary

    Parameters
    ----------
    wave : TYPE
        Description
    w0 : TYPE
        Description

    Returns
    -------
    TYPE
        Description

    """
    if isinstance(wave, Quantity):
        wave = wave.to(u.AA)
    else:
        wave = wave * u.AA
        log.info(
            "Input wavelength is not a quantity so it has no units."
            " Assuming Angstrom."
        )
    if isinstance(w0, Quantity):
        w0 = w0.to(u.AA)
    else:
        log.info(
            "Input center wavelength is not a quantity so it has no units."
            " Assuming Angstrom."
        )
        w0 = w0 * u.AA

    vel = wave.to(u.km / u.s, equivalencies=u.doppler_optical(w0))

    return vel if isinstance(wave, Quantity) else vel.value



import pyqtgraph as pg
from PyQt6 import QtGui
from PyQt6 import QtCore
import logging
import numpy as np
from astropy.units.quantity import Quantity
from astropy.io.ascii import read as ascii_read
from zhunter import DIRS
from spectres import spectres
from .colors import get_gradient
from zhunter import io

log = logging.getLogger(__name__)


class SpecSystem(QtCore.QObject):
    """
    A class to represent a spectroscopic system, either in emission
    or in absorption.
    """

    # define a custom signal to emit
    # for this to work, the object has to be a QObject
    # and need to call super().__init__() in the __init__()
    edited = QtCore.pyqtSignal()

    def __init__(
        self,
        z,
        PlotItem,
        sys_type,
        color=QtGui.QColor("cyan"),
        fname=None,
        lines=None,
        show_fs=False,
    ):
        super().__init__()
        self.redshift = z
        self.color = QtGui.QColor(color)
        self.sys_type = sys_type

        if lines is None:
            if fname is None:
                if sys_type == "em":
                    fname = DIRS["DATA"] / "lines/emission_lines.ecsv"
                elif sys_type == "abs":
                    fname = DIRS["DATA"] / "lines/basic_line_list.ecsv"
            self.lines = ascii_read(fname)
        else:
            self.lines = lines
        self.plotted_lines = []
        self.pi = PlotItem
        self.show_fs = show_fs
        self.plot_unit = None
        # Empty PlotDataItem to use for the legend entry
        self.leg_ref = pg.PlotDataItem(pen=self.color)
        if self.sys_type == "em":
            self.wave_key = "awav"
        elif self.sys_type == "abs":
            self.wave_key = "wave"
        else:
            raise ValueError

    # Used for allowing sorting
    def __lt__(self, obj):
        return (self.redshift) < (obj.redshift)

    def __gt__(self, obj):
        return (self.redshift) > (obj.redshift)

    def __le__(self, obj):
        return (self.redshift) <= (obj.redshift)

    def __ge__(self, obj):
        return (self.redshift) >= (obj.redshift)

    def __eq__(self, obj):
        return self.redshift == obj.redshift

    def draw(self, xmin=None, xmax=None, unit=None):
        """Draw the lines of the system that are within xmin and xmax.
        If no units are provided with either xmin as a Quantity or with
        the unit argument, will use the default units of the line list.

        Parameters
        ----------
        xmin : Quantity or float, optional
            Mininum range of the axis, line wavelengths below this will
            not be displayed. Also used if a Quantity and unit is None
            to infer the units of the axis.
        xmax : Quantity or float, optional
            Maxinum range of the axis, line wavelengths above this will
            not be displayed. Also used if a Quantity and unit is None
            to infer the units of the axis.
        unit : astropy Unit, optional
            Units of the axis on which to plot the lines, can be used to
            override units of xmin and xmax.
        """
        if isinstance(xmin, Quantity):
            if unit is None:
                unit = xmin.unit
            else:
                xmin = xmin.to(unit)
                xmax = xmax.to(unit)
        else:
            if unit is None:
                log.warning(
                    "No units specified through xmin as Quantity"
                    " or unit argument, using units of the line list."
                )
                unit = self.lines[self.wave_key].unit
                xmin = xmin * unit
                xmax = xmax * unit
            else:
                xmin = xmin * unit
                xmax = xmax * unit

        self.plot_unit = unit

        # Add lines to plot
        log.debug("Drawing %s system at redshift : %.5lf", self.sys_type, self.redshift)
        for w, n in zip(Quantity(self.lines[self.wave_key]), self.lines["name"]):
            if ("*" in n) and not self.show_fs:
                # If this is a fine structure line but show_fs is false, skip
                continue
            wave_obs = w * (1 + self.redshift)
            gt_min = (xmin is None) or (wave_obs >= xmin)
            lt_max = (xmax is None) or (wave_obs <= xmax)
            # If line observed wavelength is within the data range
            if gt_min and lt_max:
                if self.sys_type == "abs":
                    line = pg.InfiniteLine(
                        wave_obs.to(unit).value,
                        span=(0.0, 0.8),
                        pen=pg.mkPen(self.color, width=3),
                        hoverPen=pg.mkPen(self.color, width=6),
                        name=n,
                        label=n,
                        labelOpts={
                            "color": self.color,
                            "angle": 45,
                            "position": 1,
                        },
                        movable=True,
                    )
                elif self.sys_type == "em":
                    line = pg.InfiniteLine(
                        wave_obs.to(unit).value,
                        span=(0.2, 1.0),
                        pen=pg.mkPen(self.color, width=3),
                        hoverPen=pg.mkPen(self.color, width=6),
                        name=n,
                        label=n,
                        labelOpts={
                            "color": self.color,
                            "angle": -45,
                            "position": 0,
                        },
                        movable=True,
                    )

                line.sigPositionChangeFinished.connect(self.update_redshift)

                self.pi.addItem(line)
                self.plotted_lines.append(line)

        self.pi.legend.addItem(
            self.leg_ref,
            name=f'<p style="color:{self.color.name()};">z={self.redshift:.4f} </p>',
        )

    def undraw(self):
        self.pi.legend.removeItem(self.leg_ref)
        for line in self.plotted_lines:
            self.pi.removeItem(line)
        log.debug(
            "Removed %s system at redshift %.5lf from plot",
            self.sys_type,
            self.redshift,
        )

    def redraw(self, xmin=None, xmax=None, unit=None):
        self.undraw()
        self.draw(xmin=xmin, xmax=xmax, unit=unit)

    def update_redshift(self, line):
        """Because signal from InfiniteLine sends also the line object
        that emitted it, recover it as the second argument.
        """
        new_wave = line.getXPos() * self.plot_unit
        lname = line.name()
        rest_wave = [
            w
            for w, n in zip(self.lines[self.wave_key], self.lines["name"])
            if n in lname
        ]
        if len(rest_wave) == 1:
            rest_wave = rest_wave[0] * self.lines[self.wave_key].unit
        else:
            raise ValueError(
                f"Line name {lname} found more (or less) than one corresponding wavelength : {rest_wave}. Can't calculate a redshift."
            )
        new_z = new_wave / rest_wave - 1
        self.redshift = new_z.value
        xmin, xmax = self.pi.vb.getState()["limits"]["xLimits"]
        self.redraw(xmin=xmin, xmax=xmax, unit=self.plot_unit)
        log.info(f"Updated redshift of {self.sys_type} system to: {self.redshift:.4f}")
        self.edited.emit()


class Telluric:

    """Telluric data is in vacuum from iSpec in the optical
    (https://github.com/marblestation/iSpec)
    and Gemini Observatory in the NIR
    (https://www.gemini.edu/observing/telescopes-and-sites/sites#SkyBackground)

    Attributes
    ----------
    color : QColor
        Color used when plotting telluric spectrum
    fname : str or Path
        Name of the file containing the telluric spectrum
    vb : ViewBox
        ViewBox on which to plot the telluric spectrum
    plotted_items : list
        List of items plotted by this class
    spectrum : Spectrum1D
        Spectrum used for plotting
    spectrum_full_res : Spectrum1D
        Full resolution spectrum, saved in memory in case a lower
        resolution spectrum is plotted
    """

    def __init__(
        self,
        vb,
        color=QtGui.QColor("gray"),
        fname=DIRS["DATA"] / "tellurics/synth_tellurics_350_1100nm.csv.gz",
    ):
        self.color = color
        self.plotted_items = []
        self.vb = vb
        self.fname = fname
        self.spectrum_full_res = None

    def load_spectrum(self, fname=None, sep="\t", **args):
        if fname is None:
            fname = self.fname
        self.spectrum_full_res = io.read_generic_1D_spectrum(fname)
        self.spectrum = self.spectrum_full_res

    def draw(self, xmin=None, xmax=None):
        log.debug(f"Attempting to draw telluric spectrum from {xmin} to {xmax}")
        wave_unit = xmin.unit
        imin = self.spectrum.spectral_axis.searchsorted(xmin)
        imax = self.spectrum.spectral_axis.searchsorted(xmax)
        tellurics = pg.PlotCurveItem(
            self.spectrum.spectral_axis[imin:imax].to(wave_unit).value,
            self.spectrum.flux[imin:imax].value,
            pen=pg.mkPen(self.color, width=0.3),
            brush=QtGui.QBrush(get_gradient(QtGui.QColor(self.color))),
            fillLevel=1,
        )
        self.plotted_items.append(tellurics)
        self.vb.addItem(tellurics)

    def undraw(self):
        for item in self.plotted_items:
            self.vb.removeItem(item)

    def show(self):
        for item in self.plotted_items:
            item.show()

    def hide(self):
        for item in self.plotted_items:
            item.hide()

    def redraw(self, xmin=None, xmax=None):
        self.undraw()
        self.draw(xmin=xmin, xmax=xmax)

    def lower_resolution(self, dlam):
        """Degrade the spectrum to the desired pixel size.

        Parameters
        ----------
        dlam : astropy Quantity
            pixel size
        """
        if (self.spectrum.spectral_axis[1] - self.spectrum.spectral_axis[0]) == dlam:
            log.info(
                "The telluric specturm is already at the requested resolution. Ignoring."
            )
            return
        self.spectrum = self.spectrum_full_res.copy()
        awav_low_res = np.arange(
            self.spectrum.spectral_axis.min().value,
            self.spectrum.spectral_axis.max().value,
            dlam.value,
        )
        self.spectrum.flux = spectres(
            awav_low_res,
            self.spectrum.spectral_axis.value,
            self.spectrum.flux.value,
            fill=1,
        )
        self.spectrum.spectral_axis = awav_low_res * self.spectrum.spectral_axis.unit


class SkyBackground:

    """Sky background

    Attributes
    ----------
    color : QColor
        Color used when plotting telluric spectrum
    fname : str or Path
        Name of the file containing the telluric spectrum
    vb : ViewBox
        ViewBox on which to plot the telluric spectrum
    plotted_items : list
        List of items plotted by this class
    spectrum : Spectrum1D
        Spectrum used for plotting
    spectrum_full_res : Spectrum1D
        Full resolution spectrum, saved in memory in case a lower
        resolution spectrum is plotted
    """

    def __init__(
        self,
        vb,
        color=QtGui.QColor("gray"),
        fname=DIRS["DATA"] / "sky_background/sky_bkg_norm_nir_9000_23000.csv.gz",
    ):
        self.color = color
        self.plotted_items = []
        self.vb = vb
        self.fname = fname
        self.spectrum_full_res = None

    def load_spectrum(self, fname=None, sep="\t", **args):
        if fname is None:
            fname = self.fname
        self.spectrum_full_res = io.read_generic_1D_spectrum(fname)
        self.spectrum = self.spectrum_full_res

    def draw(self, xmin=None, xmax=None):
        log.debug(f"Attempting to draw sky background spectrum from {xmin} to {xmax}")
        wave_unit = xmin.unit
        imin = self.spectrum.spectral_axis.searchsorted(xmin)
        imax = self.spectrum.spectral_axis.searchsorted(xmax)
        tellurics = pg.PlotCurveItem(
            self.spectrum.spectral_axis[imin:imax].to(wave_unit).value,
            np.log10(self.spectrum.flux[imin:imax].value),
            pen=pg.mkPen(self.color, width=0.3),
            brush=QtGui.QBrush(get_gradient(QtGui.QColor(self.color), reverse=True)),
            fillLevel=0,
        )
        self.plotted_items.append(tellurics)
        self.vb.addItem(tellurics)

    def undraw(self):
        for item in self.plotted_items:
            self.vb.removeItem(item)

    def show(self):
        for item in self.plotted_items:
            item.show()

    def hide(self):
        for item in self.plotted_items:
            item.hide()

    def redraw(self, xmin=None, xmax=None):
        self.undraw()
        self.draw(xmin=xmin, xmax=xmax)

    def lower_resolution(self, dlam):
        """Degrade the spectrum to the desired pixel size.

        Parameters
        ----------
        dlam : astropy Quantity
            pixel size
        """
        if (self.spectrum.spectral_axis[1] - self.spectrum.spectral_axis[0]) == dlam:
            log.info(
                "The telluric specturm is already at the requested resolution. Ignoring."
            )
            return
        self.spectrum = self.spectrum_full_res.copy()
        awav_low_res = np.arange(
            self.spectrum.spectral_axis.min().value,
            self.spectrum.spectral_axis.max().value,
            dlam.value,
        )
        self.spectrum.flux = spectres(
            awav_low_res,
            self.spectrum.spectral_axis.value,
            self.spectrum.flux.value,
            fill=1,
        )
        self.spectrum.spectral_axis = awav_low_res * self.spectrum.spectral_axis.unit


class SpecSystemModel(QtCore.QAbstractListModel):
    def __init__(self, *args, specsystems=None, **kwargs):
        super(SpecSystemModel, self).__init__(*args, **kwargs)
        self.specsystems = specsystems or []

    def data(self, index, role):
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            # See below for the data structure.
            status, specsys = self.specsystems[index.row()]
            return "z = {:.5f} ({:s})".format(specsys.redshift, specsys.sys_type)

        if role == QtCore.Qt.ItemDataRole.ForegroundRole:
            status, specsys = self.specsystems[index.row()]
            return QtGui.QBrush(QtGui.QColor(specsys.color))

        if role == QtCore.Qt.ItemDataRole.DecorationRole:
            status, specsys = self.specsystems[index.row()]
            return QtGui.QColor(specsys.color)

    def sort(self):
        self.specsystems.sort(reverse=True)

    def rowCount(self, index):
        return len(self.specsystems)

    def delete(self, index):
        # Remove the item and refresh.
        _, _sys = self.specsystems[index.row()]
        log.debug("Received request to delete system at redshift %.5lf", _sys.redshift)
        _sys.undraw()
        del self.specsystems[index.row()]
        self.layoutChanged.emit()
        self.sort()

    def clear(self):
        for _, specsys in self.specsystems:
            specsys.undraw()
        self.specsystems = []

    def get_color(self, index):
        _, _sys = self.specsystems[index.row()]
        return _sys.color

    def get_redshift(self, index):
        _, _sys = self.specsystems[index.row()]
        return _sys.redshift

    def get_specsys(self, index):
        _, _sys = self.specsystems[index.row()]
        return _sys

    def show_hide_fine_structure(self, index, bounds):
        _, _sys = self.specsystems[index.row()]
        fs_shown = _sys.show_fs
        _sys.show_fs = not fs_shown
        _sys.redraw(xmin=bounds[0], xmax=bounds[1])
        if fs_shown:
            log.debug(
                "Hiding fine structure for system at redshift %.5lf", _sys.redshift
            )
        else:
            log.debug(
                "Showing fine structure for system at redshift %.5lf", _sys.redshift
            )



from itertools import cycle
import logging
import sys
from pathlib import Path

from PyQt6 import uic
from PyQt6 import QtWidgets
from PyQt6 import QtCore
from PyQt6 import QtGui

from astropy.io import fits
from astropy.io.ascii import read as ascii_read
from astropy.units.quantity import Quantity
import astropy.constants as cst
from astropy.table import join

import numpy as np
import pyqtgraph as pg
from zhunter import DIRS
import zhunter.io as io
import zhunter.spectral_functions as sf
from .spectroscopic_system import SpecSystem, SpecSystemModel, Telluric, SkyBackground
from .line_list_selection import SelectLineListsDialog, select_file
from .velocity_plot import VelocityPlot
from .key_binding import KeyBindingHelpDialog
from .misc import create_line_ratios, set_up_linked_vb, check_flux_scale
from .colors import COLORS

logging.getLogger("PyQt6").setLevel(logging.INFO)
logging.getLogger("matplotlib").setLevel(logging.INFO)
log = logging.getLogger(__name__)
logging.basicConfig(
    stream=sys.stdout,
    level=logging.DEBUG,
    format="%(asctime)s %(levelname)s [%(name)s] %(message)s",
)


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super(MainWindow, self).__init__(*args, **kwargs)

        # Define color style
        # Have to do this before loading UI for it to work
        # self.color_style = "old"
        # self.color_style = "cyberpunk"
        self.color_style = "kraken"
        # self.color_style = "cvd"
        self.colors = COLORS[self.color_style]
        pg.setConfigOption("foreground", self.colors["foreground"])
        pg.setConfigOption("background", self.colors["background"])

        # Load the UI Page
        # pg.setConfigOption("foreground", "w")
        uic.loadUi(DIRS["UI"] / "main_frame.ui", self)
        self.statusBar = QtWidgets.QStatusBar()
        self.setStatusBar(self.statusBar)

        # Load the line lists
        self.fnames = {}
        self.fnames["data"] = None
        self.fnames["emission_lines"] = DIRS["DATA"] / "lines/emission_lines.ecsv"
        self.fnames["absorption_lines"] = DIRS["DATA"] / "lines/basic_line_list.ecsv"
        self.fnames["fine_structure_lines"] = DIRS["DATA"] / "lines/fine_structure.ecsv"
        self.fnames["line_ratio"] = DIRS["DATA"] / "lines/line_ratio.csv"
        self.fnames["tellurics"] = (
            DIRS["DATA"] / "tellurics/synth_tellurics_350_1100nm.csv.gz"
        )
        self.fnames["sky_bkg"] = (
            DIRS["DATA"] / "sky_background/sky_bkg_norm_nir_9000_23000.csv.gz"
        )
        self.load_line_lists(calc_ratio=False)

        # List of spectral systems
        self.specsysModel = SpecSystemModel()
        self.specsysView.setModel(self.specsysModel)
        self.specsysView.setStyleSheet(
            f"QListView{{background-color: {self.colors['background']};}}"
        )
        # Signals and slots of spectral systems
        self.add_abs_button.clicked.connect(self.add_absorber)
        self.add_em_button.clicked.connect(self.add_emitter)
        self.del_specsys_button.clicked.connect(self.delete_specsys)
        self.file_1D_button.clicked.connect(self.select_1D_file)
        self.file_2D_button.clicked.connect(self.select_2D_file)
        self.file_line_list_button.clicked.connect(self.select_line_lists)
        self.velocity_plot_button.clicked.connect(self.velocity_plot)

        # Help
        self.actionKey_Bindings.triggered.connect(KeyBindingHelpDialog(self).show)
        self.actionKey_Bindings.setShortcut(QtCore.Qt.Key.Key_H)

        # Connect all signals and slots
        self.show_uncertainty_cb.stateChanged.connect(self.show_hide_uncertainty)
        self.telluric_cb.stateChanged.connect(self.show_hide_telluric)
        self.sky_bkg_cb.stateChanged.connect(self.show_hide_sky_bkg)
        self.fine_structure_button.clicked.connect(self.show_hide_fine_structure)
        self.to_vacuum_button.clicked.connect(self.wvlg_to_vacuum)
        self.to_air_button.clicked.connect(self.wvlg_to_air)
        self.actionBarycentric.triggered.connect(self.wvlg_bary_correction)
        self.actionHeliocentric.triggered.connect(self.wvlg_helio_correction)
        self.ratio_button.clicked.connect(self.calculate_ratio)
        self.find_line_ratios_button.clicked.connect(self.find_ratio_names)
        self.add_ratio_button.clicked.connect(self.add_specsys_from_ratio)
        self.add_line_button.clicked.connect(self.add_specsys_from_line)
        self.reset_smooth_button.clicked.connect(self.reset_smoothing)
        self.textbox_for_smooth.editingFinished.connect(self.apply_smoothing)
        self.reset_width_button.clicked.connect(self.reset_width)
        self.textbox_for_extraction_width.editingFinished.connect(
            self.set_extraction_width
        )
        self.graphLayout.setFocus()

        # Call reset to set general properties that will be used
        self.reset_plot()

    # Setting up of plotting items
    def set_up_plot(self):
        if self.mode == "1D":
            # Add title
            self.graphLayout.addLabel(self.fnames["data"].name, row=0)
            # Define PlotItem as ax1D (subclass of GraphicsItem) on wich to plot stuff
            self.ax1D = self.graphLayout.addPlot(row=1, col=0)

        elif self.mode == "2D":
            # Add title on row 0
            self.graphLayout.addLabel(self.fnames["data"].name, row=0, colspan=2)
            # Define PlotItem as ax1D and ax2D (subclass of GraphicsItem) on wich to plot stuff
            self.ax2D = self.graphLayout.addPlot(row=1, col=0)
            self.ax1D = self.graphLayout.addPlot(row=2, col=0)
            self.ax2D_side_vb = self.graphLayout.addViewBox(row=1, col=1)
            self.img_colorbar = pg.HistogramLUTItem(gradientPosition="left")
            self.graphLayout.addItem(self.img_colorbar, row=2, col=1)
            # Strech rows to make 1D bigger than 2D plot in y direction
            self.graphLayout.ci.layout.setRowStretchFactor(1, 2)
            self.graphLayout.ci.layout.setRowStretchFactor(2, 3)
            # Strech column 0 (where 1D and 2D plots are) to make it bigger in x than the side histograms
            self.graphLayout.ci.layout.setColumnStretchFactor(0, 100)

            # Remove padding so that panning with keyboard preserves x and y range
            self.ax2D.vb.setDefaultPadding(padding=0.00)

            # Link the side histograms to the central plots
            self.ax1D.vb.setXLink(self.ax2D.vb)
            self.ax2D_side_vb.setYLink(self.ax2D.vb)

            # Hide the axis to make the plot prettier and more compact
            self.ax2D.hideAxis("bottom")

            # Change all the widths to be equal so things are aligned
            self.ax2D.getAxis("left").setWidth(60)

            # Remove mouse interactions on side histogram
            self.ax2D_side_vb.setMouseEnabled(x=False, y=True)

            # Add the PlotItems to the list of active items
            # self.active_plot_items.append(self.ax2D)
            # self.active_plot_items.append(self.ax2D_side_vb)

        # Set up the legend (empty for now)
        legend = pg.LegendItem(
            offset=(-10, 1),
            colCount=2,
            horSpacing=20,
            verSpacing=-5,
        )
        legend.setParentItem(self.ax1D)
        self.ax1D.legend = legend

        # Fix the size of left axis so the center panels align vertically
        self.ax1D.getAxis("left").setWidth(60)
        self.ax1D.getAxis("bottom").setHeight(30)
        # self.active_plot_items.append(self.ax1D)

        # Remove padding so that panning preserves x and y range
        self.ax1D.vb.setDefaultPadding(padding=0.00)

        # cross hair for 1D plot
        self.set_up_crosshairs()

        # ViewBox for Tellurics
        self.telluric_vb = set_up_linked_vb(self.ax1D)
        # ViewBox for sky background
        self.sky_bkg_vb = set_up_linked_vb(self.ax1D)

        # To catch the key presses from the PlotItem
        self.ax1D.installEventFilter(self)
        self.ax1D.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        if self.mode == "2D":
            # Install also on 2D PlotItem
            self.ax2D.installEventFilter(self)
            self.ax2D.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

        # Create empty objects that will hold the data to be displayed
        self.__create_placeholders()
        self.__reset_colors()

    def set_up_crosshairs(self):
        """
        Set up the crosshairs to be displayed on both 1D and 2D
        """
        self.crosshair_x_1D = pg.InfiniteLine(
            angle=90, movable=False, pen=pg.mkPen(color=self.colors["crosshair"])
        )
        self.crosshair_y_1D = pg.InfiniteLine(
            angle=0, movable=False, pen=pg.mkPen(color=self.colors["crosshair"])
        )
        self.ax1D.addItem(self.crosshair_x_1D, ignoreBounds=True)
        self.ax1D.addItem(self.crosshair_y_1D, ignoreBounds=True)
        self.ax1D.scene().sigMouseMoved.connect(self.move_crosshair)
        if self.mode == "2D":
            self.crosshair_x_2D = pg.InfiniteLine(
                angle=90, movable=False, pen=pg.mkPen(color=self.colors["crosshair"])
            )
            self.crosshair_y_2D = pg.InfiniteLine(
                angle=0, movable=False, pen=pg.mkPen(color=self.colors["crosshair"])
            )
            self.crosshair_y_2D_side = pg.InfiniteLine(
                angle=0, movable=False, pen=pg.mkPen(color=self.colors["crosshair"])
            )
            self.ax2D.addItem(self.crosshair_x_2D, ignoreBounds=True)
            self.ax2D.addItem(self.crosshair_y_2D, ignoreBounds=True)
            self.ax2D_side_vb.addItem(self.crosshair_y_2D_side, ignoreBounds=True)
            self.crosshair_x_2D.setZValue(9)  # To make sure crosshair is on top
            self.crosshair_y_2D.setZValue(9)  # To make sure crosshair is on top
            self.ax2D.scene().sigMouseMoved.connect(self.move_crosshair)
            self.ax2D_side_vb.scene().sigMouseMoved.connect(self.move_crosshair)

    def set_up_img_colobar(self):
        self.img_colorbar.setImageItem(self.flux_2D_img)
        self.img_colorbar.setHistogramRange(
            self.data["q025_2D"].value, self.data["q975_2D"].value
        )
        self.img_colorbar.setLevels(
            self.data["q025_2D"].value, self.data["q975_2D"].value
        )
        cmap = pg.colormap.get("afmhot", source="matplotlib")
        self.img_colorbar.gradient.setColorMap(cmap)

    def set_up_ROI(self):
        spatial_width = float(self.textbox_for_extraction_width.text())
        self.roi = pg.ROI(
            pos=[
                self.data["wvlg_min"].value,
                self.data["spat_med"].value - spatial_width / 2,
            ],
            size=[self.data["wvlg_span"].value, spatial_width],
            pen=pg.mkPen(self.colors["roi"], width=2),
            hoverPen=pg.mkPen(self.colors["roi"], width=5),
            handlePen=pg.mkPen(self.colors["roi"], width=2),
            handleHoverPen=pg.mkPen(self.colors["roi"], width=5),
        )
        self.ax2D.addItem(self.roi)
        self.roi.setZValue(10)
        # Extend ROI visualization to side histogram
        self.ROI_y_hist_lower = pg.InfiniteLine(
            self.roi.pos()[1],
            angle=0,
            movable=False,
            pen=pg.mkPen(self.colors["roi"], width=2),
        )

        self.ROI_y_hist_upper = pg.InfiniteLine(
            self.roi.pos()[1] + self.roi.size()[1],
            angle=0,
            movable=False,
            pen=pg.mkPen(self.colors["roi"], width=2),
        )

        self.ax2D_side_vb.addItem(self.ROI_y_hist_lower, ignoreBounds=True)
        self.ax2D_side_vb.addItem(self.ROI_y_hist_upper, ignoreBounds=True)
        self.roi.sigRegionChanged.connect(self.extract_and_plot_1D)

    def set_labels(self):
        self.ax1D.setLabel(
            "left",
            "Flux" + f" ({self.data['flux_1D'].unit})",
            # useful if you want pyqtgraph to automatically display k in front
            # of units if you zoom out to thousands for example
            # units=f"{self.data['flux_1D'].unit}",
        )
        self.ax1D.setLabel(
            "bottom",
            "Observed wavelength" + f" ({self.data['wvlg'].unit})",
        )
        # self.ax1D.showGrid(x=True, y=True)
        if self.mode == "2D":
            self.ax2D.setLabel(
                "left",
                "Spatial" + f" ({self.data['spat'].unit})",
            )

    def set_ViewBox_limits(self):
        self.ax1D.vb.setLimits(
            xMin=self.data["wvlg_min"].value, xMax=self.data["wvlg_max"].value
        )
        if self.mode == "2D":
            self.ax2D.vb.setLimits(
                xMin=self.data["wvlg_min"].value,
                xMax=self.data["wvlg_max"].value,
                yMin=self.data["spat_min"].value,
                yMax=self.data["spat_max"].value,
            )
            self.ax2D_side_vb.setLimits(
                yMin=self.data["spat_min"].value, yMax=self.data["spat_max"].value
            )

    def __create_placeholders(self):
        """
        Create empty objects that will hold the 1D and 2D data
        to be displayed.
        """
        self.flux_1D_spec = pg.PlotCurveItem(
            np.zeros(2),
            np.zeros(1),
            stepMode="center",
            pen=pg.mkPen(color=self.colors["spec"]),
        )
        self.unc_1D_spec = pg.PlotCurveItem(
            np.zeros(2),
            np.zeros(1),
            stepMode="center",
            pen=pg.mkPen(color=self.colors["unc"]),
        )
        self.lam1_line = pg.InfiniteLine(
            0,
            span=(0.9, 1.0),
            pen=pg.mkPen(
                color=self.colors["foreground"],
                width=2,
                style=QtCore.Qt.PenStyle.DashLine,
            ),
            label="Lam1",
            labelOpts={
                "color": QtGui.QColor(self.colors["foreground"]),
                "position": 0.5,
            },
        )
        self.lam2_line = pg.InfiniteLine(
            0,
            span=(0.9, 1.0),
            pen=pg.mkPen(
                color=self.colors["foreground"],
                width=2,
                style=QtCore.Qt.PenStyle.DashLine,
            ),
            label="Lam2",
            labelOpts={
                "color": QtGui.QColor(self.colors["foreground"]),
                "position": 0.5,
            },
        )
        self.telluric_1D_spec = None
        self.sky_bkg_1D_spec = None
        self.ax1D.addItem(self.flux_1D_spec)
        self.ax1D.addItem(self.unc_1D_spec)
        self.ax1D.addItem(self.lam1_line)
        self.ax1D.addItem(self.lam2_line)

        if self.mode == "2D":
            self.flux_2D_img = pg.ImageItem()
            # Monkey-patch the image to use our custom hover function.
            # This is generally discouraged (I should subclass ImageItem instead),
            # but it works for a very simple use like this.
            self.flux_2D_img.hoverEvent = self.image_hover_event
            self.unc_2D_img = pg.ImageItem()
            self.sidehist_2D = pg.PlotCurveItem(
                np.zeros(1), np.zeros(1), pen=pg.mkPen(color=self.colors["foreground"])
            )
            self.ax2D.addItem(self.flux_2D_img)
            self.ax2D.addItem(self.unc_2D_img)
            self.ax2D_side_vb.addItem(self.sidehist_2D)

    # Colors
    def __reset_colors(self):
        """
        Reset the color palet.
        """
        self.available_colors = self.colors["specsys"].copy()
        self.available_colors_cycler = cycle(self.available_colors)

    def get_color(self):
        """
        Get the next color from the list of available colors.
        If all colors have been used, reset the color cycler.
        """
        try:
            color = next(self.available_colors_cycler)
        except StopIteration:
            log.info("Exhausted all colors, resetting color cycler.")
            self.__reset_colors()
            color = next(self.available_colors_cycler)
        log.debug("There are %d unused colors left", len(self.available_colors))
        return color

    def clear_color_from_available_list(self, color):
        """
        Remove the color from the pool of available colors.
        """
        self.available_colors.remove(color)
        self.available_colors_cycler = cycle(self.available_colors)

    def visualize_spec(self):
        """
        Main function to be called once to set up the visualization.
        It does, in this order :
            1) Read data
            2) Load in into memory
            3) Show the data on the interface
            4)  a) Set up ROI for 1D extraction if 2D mode
                b) Extract, load into memory and show 1D
            5) Finalize display by adding labels, etc.
        """
        # Read data
        try:
            if self.mode == "1D":
                spec_1D = io.read_1D_data(self.fnames["data"])
                wvlg_1D = spec_1D.spectral_axis
                flux_1D = spec_1D.flux

                if spec_1D.uncertainty:
                    # Uncertainty is not a Quantity but a StdDevUncertainty object
                    unc_1D = spec_1D.uncertainty.array * flux_1D.unit
                else:
                    log.warning(
                        f"No uncertainty/error spectrum found in {self.fnames['data']}, using 0."
                    )
                    unc_1D = np.zeros(wvlg_1D.shape) * flux_1D.unit
            elif self.mode == "2D":
                wvlg, spat, flux, unc = io.read_fits_2D_spectrum(self.fnames["data"])
        except Exception as e:
            log.error(e)
            QtWidgets.QMessageBox.information(self, "Invalid input file", str(e))
            self.reset_plot()
            return

        if ".fit" in str(self.fnames["data"].stem):
            try:
                self.data["header"] = fits.getheader(self.fnames["data"])
            except OSError:
                log.warning(
                    f"No header could be loaded for file {self.fnames['data']}."
                )
                self.data["header"] = None
        else:
            log.info("Not a fits file, no header loaded.")
            self.data["header"] = None

        if self.mode == "1D":
            # Load data into memory
            self.load_1D_data(wvlg_1D, flux_1D, unc_1D)
        elif self.mode == "2D":
            # Load data into memory
            self.load_2D_data(wvlg, spat, flux, unc)
            # Region of Interest
            self.set_up_ROI()

        self.draw_data()

        self.ax1D.setYRange(
            min=self.data["q025_1D"].value,
            max=self.data["q975_1D"].value,
        )

        # Create appropriate labels
        self.set_labels()

        # Set the zooming limits
        self.set_ViewBox_limits()

        # Plot telluric
        if self.telluric_cb.isChecked():
            self.plot_telluric()
        # Plot sky background
        if self.sky_bkg_cb.isChecked():
            self.plot_sky_bkg()

    def load_1D_data(self, wvlg, flux, unc):
        """
        Save wvlg, flux and unc arrays into memory.
        """
        # Multiply flux and unc to have them in reasonable units
        # and allow y crosshair to work (otherwise the code considers
        # it to be zero)
        flux, unc = check_flux_scale(flux, unc)
        self.data["wvlg"] = wvlg
        self.data["wvlg_1D"] = wvlg
        self.data["flux_1D"] = flux
        self.data["unc_1D"] = unc

        self.set_1D_displayed_data(
            wvlg=self.data["wvlg_1D"],
            flux=self.data["flux_1D"],
            unc=self.data["unc_1D"],
        )
        self.calculate_1D_displayed_data_range()

    def load_2D_data(self, wvlg, spat, flux, unc):
        """
        Loads wvlg, spat, flux and unc arrays into memory.
        """
        # Multiply flux and unc to have them in reasonable units
        # and allow y crosshair to work (otherwise the code considers
        # it to be zero)
        flux, unc = check_flux_scale(flux, unc)

        self.data["wvlg"] = wvlg
        self.data["flux_2D"] = flux
        self.data["unc_2D"] = unc
        self.data["spat"] = spat

        self.set_2D_displayed_data(
            wvlg=self.data["wvlg"],
            flux=self.data["flux_2D"],
            unc=self.data["unc_2D"],
            spat=self.data["spat"],
        )
        self.calculate_2D_displayed_data_range()

    def load_line_lists(self, calc_ratio):
        """
        Load the input line lists and check the format is ok.
        """
        self.abs_lines = ascii_read(self.fnames["absorption_lines"])
        log.debug(f"Read absorption lines from: {self.fnames['absorption_lines']}")

        self.em_lines = ascii_read(self.fnames["emission_lines"])
        log.debug(f"Read emission lines from: {self.fnames['emission_lines']}")

        self.fs_lines = ascii_read(self.fnames["fine_structure_lines"])
        check = (
            "name" not in self.abs_lines.columns
            or "wave" not in self.abs_lines.columns
            or "name" not in self.em_lines.columns
            or "awav" not in self.em_lines.columns
            or "name" not in self.fs_lines.columns
            or "wave" not in self.fs_lines.columns
        )
        if check:
            QtWidgets.QMessageBox.information(
                self,
                "Invalid line format",
                "Column 'name' and 'wave' must exist. "
                "Please specify them as the first uncommented "
                "line of csv your file.",
            )
            return

        if calc_ratio:
            self.line_ratios = create_line_ratios(self.fnames["absorption_lines"])
        else:
            self.line_ratios = ascii_read(self.fnames["line_ratio"])
            log.debug(f"Read line ratios: {self.fnames['line_ratio']}")

    def set_2D_displayed_data(self, wvlg, flux, unc, spat):
        """
        Saves into memory the data that is actually being displayed
        on the interface (after smoothing for example).
        """

        self.data["wvlg_2D_disp"] = wvlg
        self.data["flux_2D_disp"] = flux
        self.data["unc_2D_disp"] = unc
        self.data["spat_disp"] = spat

    def set_1D_displayed_data(self, wvlg, flux, unc):
        """
        Saves into memory the data that is actually being displayed
        on the interface (after smoothing for example).
        """
        # Modify wvlg array to be of size len(flux)+1 by adding the right
        # edge of the last bin. This is to allow for visualization with
        # stepMode='center'
        wvlg_unit = wvlg.unit
        wvlg = wvlg.value
        dx = wvlg[1] - wvlg[0]
        # Add the right edge of the final bin
        self.data["wvlg_1D_disp"] = np.asarray(list(wvlg) + [wvlg[-1] + dx]) * wvlg_unit
        self.data["flux_1D_disp"] = flux
        self.data["unc_1D_disp"] = unc

    def calculate_1D_displayed_data_range(self):
        """
        Compute the range spanned by the displayed data for
        visualization purposed such as setting the min and max range
        allowed by the ViewBox.
        """
        self.data["wvlg_min"] = np.min(self.data["wvlg_1D_disp"])
        self.data["wvlg_max"] = np.max(self.data["wvlg_1D_disp"])
        self.data["wvlg_span"] = self.data["wvlg_max"] - self.data["wvlg_min"]
        self.data["q975_1D"] = np.quantile(self.data["flux_1D_disp"], q=0.975)
        self.data["q025_1D"] = np.quantile(self.data["flux_1D_disp"], q=0.025)

    def calculate_2D_displayed_data_range(self):
        """
        Compute the range spanned by the displayed data for
        visualization purposed such as setting the min and max range
        allowed by the ViewBox.
        """
        self.data["wvlg_min"] = np.min(self.data["wvlg_2D_disp"])
        self.data["wvlg_max"] = np.max(self.data["wvlg_2D_disp"])
        self.data["wvlg_span"] = self.data["wvlg_max"] - self.data["wvlg_min"]
        self.data["q975_2D"] = np.quantile(self.data["flux_2D_disp"], q=0.975)
        self.data["q025_2D"] = np.quantile(self.data["flux_2D_disp"], q=0.025)
        self.data["spat_min"] = np.min(self.data["spat_disp"])
        self.data["spat_max"] = np.max(self.data["spat_disp"])
        self.data["spat_med"] = np.median(self.data["spat_disp"])
        self.data["spat_span"] = np.max(self.data["spat_disp"]) - np.min(
            self.data["spat_disp"]
        )

    def draw_data(self):
        """
        A wrapper function to draw data. Look at draw_1D_data and
        draw_2D_data for more details.
        """
        if self.mode == "2D":
            self.draw_2D_data()
            self.extract_and_plot_1D()
        elif self.mode == "1D":
            self.draw_1D_data()
        else:
            log.error("This should never happend. Should always be '1D' or '2D' mode.")

    def draw_2D_data(self):
        """
        Takes the 2D display data loaded and plots it on the interface.
        """

        # Use the transpose here so that the wavelength and spatial dimensions are in the right order
        self.flux_2D_img.setImage(
            self.data["flux_2D_disp"].T.value,
            levels=(self.data["q025_2D"].value, self.data["q975_2D"].value),
        )

        # This is will not be seen but is needed when doing the ROI extraction
        # Essentially, we're overlaying 2 images, one of the flux and one of the errors
        self.unc_2D_img.setImage(self.data["unc_2D_disp"].T.value)

        # Transform image indexes to physical coordinates
        self.rect = QtCore.QRectF(
            self.data["wvlg_2D_disp"][0].value,
            self.data["spat_disp"][0].value,
            self.data["wvlg_2D_disp"][-1].value - self.data["wvlg_2D_disp"][0].value,
            self.data["spat_disp"][-1].value - self.data["spat_disp"][0].value,
        )
        self.roi.maxBounds = self.rect
        self.flux_2D_img.setRect(self.rect)
        self.unc_2D_img.setRect(self.rect)
        self.flux_2D_img.setZValue(8)
        self.unc_2D_img.setZValue(7)
        # Don't display uncertainty image, just keep it for when extracting the 1D from the 2D
        self.unc_2D_img.hide()

        # Add the side histogram of the pixel intensities
        self.set_up_img_colobar()
        # Add the side histogram fo the flux values as a function of spatial position
        self.plot_2D_sidehist()

    def draw_1D_data(self):
        """
        Takes the 1D display data loaded and plots it on the interface.
        """
        self.flux_1D_spec.setData(
            self.data["wvlg_1D_disp"].value, self.data["flux_1D_disp"].value
        )
        self.unc_1D_spec.setData(
            self.data["wvlg_1D_disp"].value, self.data["unc_1D_disp"].value
        )

    def plot_2D_sidehist(self):
        y_dist = np.median(self.data["flux_2D_disp"], axis=1)
        self.sidehist_2D.setData(y_dist.value, self.data["spat_disp"].value)

    def plot_telluric(self):
        self.telluric_1D_spec = Telluric(vb=self.telluric_vb, color=self.colors["sky"])
        self.telluric_1D_spec.load_spectrum()
        self.telluric_1D_spec.draw(
            xmin=self.data["wvlg_min"], xmax=self.data["wvlg_max"]
        )

    def plot_sky_bkg(self):
        self.sky_bkg_1D_spec = SkyBackground(
            vb=self.sky_bkg_vb, color=self.colors["sky"]
        )
        self.sky_bkg_1D_spec.load_spectrum()
        self.sky_bkg_1D_spec.draw(
            xmin=self.data["wvlg_min"], xmax=self.data["wvlg_max"]
        )

    def show_hide_uncertainty(self):
        """
        Show or hide the 1D uncertaintyspectrum.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        if self.show_uncertainty_cb.isChecked():
            self.unc_1D_spec.show()
        else:
            self.unc_1D_spec.hide()

    def show_hide_telluric(self):
        """
        Show or hide telluric absorption on the 1D spectrum.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        if self.telluric_1D_spec is None:
            self.plot_telluric()
        if self.telluric_cb.isChecked():
            self.telluric_1D_spec.show()
        else:
            self.telluric_1D_spec.hide()

    def show_hide_sky_bkg(self):
        """
        Show or hide sky background emission on the 1D spectrum.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        if self.sky_bkg_1D_spec is None:
            self.plot_sky_bkg()
        if self.sky_bkg_cb.isChecked():
            self.sky_bkg_1D_spec.show()
        else:
            self.sky_bkg_1D_spec.hide()

    def show_hide_fine_structure(self):
        """
        Show or hide fine structure lines for the selected
        spectroscopic system
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        indexes = self.specsysView.selectedIndexes()
        if indexes:
            # Indexes is a list of a single item in single-select mode.
            index = indexes[0]
            self.specsysModel.show_hide_fine_structure(
                index, bounds=[self.data["wvlg_min"], self.data["wvlg_max"]]
            )
        else:
            QtWidgets.QMessageBox.information(
                self,
                "No spectroscopic system selected",
                "Please select a spectroscopic system from the list "
                "to show/hide the fine structure lines.",
            )

    # Events
    def eventFilter(self, widget, event):
        """
        General event catcher for keypresses.
        """
        if event.type() == QtCore.QEvent.Type.KeyPress:
            if widget is self.ax1D:
                crosshair = self.crosshair_x_1D
            elif widget is self.ax2D:
                crosshair = self.crosshair_x_2D
            else:
                return

            vb = widget.vb
            key = event.key()
            x_pos = crosshair.getPos()[0]
            x_view, y_view = vb.getState()["viewRange"]

            # Setting lambda 1 and 2
            if key == QtCore.Qt.Key.Key_Q:
                self.statusBar.showMessage(f"Setting Lambda_1 at {x_pos:0.5f} AA", 2000)
                self.textbox_for_wvlg1.setText("{:.5f}".format(x_pos))
                self.lam1_line.setPos(x_pos)
            elif key == QtCore.Qt.Key.Key_E:
                self.statusBar.showMessage(f"Setting Lambda_2 at {x_pos:0.5f} AA", 2000)
                self.textbox_for_wvlg2.setText("{:.5f}".format(x_pos))
                self.lam2_line.setPos(x_pos)
            # Panning with keyboard
            # The value returned after setting the range is slightly
            # larger (because of padding) and this results in 'zooming out'
            # after multiple key presses... Had to force padding to 0 when
            # defining the viewBox to remove this effect
            elif key == QtCore.Qt.Key.Key_D:
                vb.setRange(
                    xRange=np.array(x_view) + 0.15 * np.abs(x_view[1] - x_view[0])
                )
            elif key == QtCore.Qt.Key.Key_A:
                vb.setRange(
                    xRange=np.array(x_view) - 0.15 * np.abs(x_view[1] - x_view[0])
                )
            elif key == QtCore.Qt.Key.Key_W:
                vb.setRange(
                    yRange=np.array(y_view) + 0.15 * np.abs(y_view[1] - y_view[0])
                )
            elif key == QtCore.Qt.Key.Key_S:
                vb.setRange(
                    yRange=np.array(y_view) - 0.15 * np.abs(y_view[1] - y_view[0])
                )
        return QtWidgets.QWidget.eventFilter(self, widget, event)

    def move_crosshair(self, evt):
        pos = evt
        # Move 1D crosshair
        if self.ax1D.sceneBoundingRect().contains(pos):
            mousePoint = self.ax1D.vb.mapSceneToView(pos)
            # Avoid the case where wvlg_1D is not defined because something crashed
            if self.data.get("wvlg_1D") is not None:
                self.statusBar.showMessage(
                    f"Wavelength = {mousePoint.x():0.3f} {self.data['wvlg_1D'].unit}, "
                    f"Flux = {mousePoint.y():0.3f} {self.data['flux_1D'].unit}"
                )
            self.crosshair_x_1D.setPos(mousePoint.x())
            self.crosshair_y_1D.setPos(mousePoint.y())
            if self.mode == "2D":
                self.crosshair_x_2D.setPos(mousePoint.x())
        # If 2D mode, check for movement on 2D ax
        if self.mode == "2D":
            if self.ax2D.sceneBoundingRect().contains(pos):
                mousePoint = self.ax2D.vb.mapSceneToView(pos)
                self.crosshair_x_2D.setPos(mousePoint.x())
                self.crosshair_y_2D.setPos(mousePoint.y())
                self.crosshair_x_1D.setPos(mousePoint.x())
                self.crosshair_y_2D_side.setPos(mousePoint.y())
            elif self.ax2D_side_vb.sceneBoundingRect().contains(pos):
                mousePoint = self.ax2D_side_vb.mapSceneToView(pos)
                self.statusBar.showMessage(
                    f"Spatial = {mousePoint.y():0.3f} {self.data['spat'].unit}"
                )
                self.crosshair_y_2D.setPos(mousePoint.y())
                self.crosshair_y_2D_side.setPos(mousePoint.y())

    def image_hover_event(self, event):
        """
        Show the position and value under the mouse cursor.
        """
        if event.isExit():
            self.statusBar.showMessage("")
            return
        pos = event.pos()
        # x and y are reversed because the transpose of the flux is displayed
        i, j = pos.y(), pos.x()
        # Clip indexes to be 0 at minimum and len(flux)-1 at maximum
        i = int(np.clip(i, 0, self.data["flux_2D_disp"].shape[0] - 1))
        j = int(np.clip(j, 0, self.data["flux_2D_disp"].shape[1] - 1))
        z = self.data["flux_2D_disp"][i, j]
        x, y = self.data["wvlg"][j], self.data["spat_disp"][i]
        self.statusBar.showMessage(
            f"Wavelength = {x.value:0.3f} {x.unit},"
            f" Spatial = {y.value:0.3f} {y.unit},"
            f" Flux = {z.value:.4f} {z.unit}"
        )

    # ROI
    def show_ROI_on_sidehist(self):
        """
        Show the extension of the Region Of Interest (ROI) on the
        side histogram next to the 2D plot.
        """
        self.ROI_y_hist_lower.setPos(self.roi.pos()[1])
        self.ROI_y_hist_upper.setPos(self.roi.pos()[1] + self.roi.size()[1])

    def extract_and_plot_1D(self):
        """
        Extract the 2D data within the ROI as 1D data, load it into
        memory, apply smoothing if necessary and draw it on the 1D
        plot.
        """
        self.show_ROI_on_sidehist()
        wvlg, flux, unc = self.get_data_from_ROI()
        self.load_1D_data(wvlg, flux, unc)
        if int(self.textbox_for_smooth.text()) != 1:
            wvlg, flux, unc = self.smooth()
            self.set_1D_displayed_data(wvlg, flux, unc)
        self.draw_1D_data()

    def get_data_from_ROI(self):
        """
        Return the mean of the flux and uncertaintyin the area selected
        by the Region Of Interest widget.
        """
        flux_selected = self.roi.getArrayRegion(
            self.data["flux_2D_disp"].T.value, self.flux_2D_img, returnMappedCoords=True
        )

        unc_selected = self.roi.getArrayRegion(
            self.data["unc_2D_disp"].T.value, self.unc_2D_img, returnMappedCoords=True
        )

        flux_1D = flux_selected[0].sum(axis=1) * self.data["flux_2D"].unit
        unc_1D = np.sqrt((unc_selected[0] ** 2).sum(axis=1)) * self.data["flux_2D"].unit
        wvlg_1D = flux_selected[1][0, :, 0] * self.data["wvlg"].unit

        return wvlg_1D, flux_1D, unc_1D

    def set_extraction_width(self):
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        if self.mode == "2D":
            try:
                ext_width = float(self.textbox_for_extraction_width.text())
                if ext_width >= self.data["spat_span"].value:
                    QtWidgets.QMessageBox.information(
                        self,
                        "Invalid extraction width",
                        "Can't change extraction width: it must be "
                        "smaller than the spatial width spanned by the "
                        "spectrum",
                    )
                    return
                self.roi.setSize([self.data["wvlg_span"].value, ext_width])
            except ValueError:
                QtWidgets.QMessageBox.information(
                    self,
                    "Invalid extraction width",
                    "Can't change extraction width: it must be " "convertible to float",
                )

    def reset_width(self):
        """
        Reset the ROI region's position and extraction width to
        default values.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        if self.mode == "2D":
            self.ax2D_side_vb.removeItem(self.ROI_y_hist_lower)
            self.ax2D_side_vb.removeItem(self.ROI_y_hist_upper)
            self.ax2D.removeItem(self.roi)
            self.textbox_for_extraction_width.setText("1")
            self.set_up_ROI()
            self.extract_and_plot_1D()
        elif self.mode == "1D":
            QtWidgets.QMessageBox.information(
                self,
                "Wrong mode",
                "This feature can only be used in 2D mode.",
            )

    # File selection
    def select_1D_file(self):
        self.mode = "1D"
        log.info("Starting 1D mode!")
        self.select_file_and_plot()

    def select_2D_file(self):
        self.mode = "2D"
        log.info("Starting 2D mode!")
        self.select_file_and_plot()

    def select_file_and_plot(self):
        fname = select_file(
            self, self.fnames["data"], file_type="(*.fits *.dat *.txt *.csv *.gz)"
        )
        if fname:
            self.fnames["data"] = Path(fname)
            if self.fnames["data"].exists():
                self.reset_plot()
                self.set_up_plot()
                self.visualize_spec()

    def reset_plot(self):
        """
        Clear the plot, reset the data in memory, the various
        corrections and the list of spectroscopic systems.
        """

        # Clear the individual active plot items
        # try:
        #     for item in self.active_plot_items:
        #         item.clear()
        # except AttributeError:
        #     pass
        # self.active_plot_items = []

        # Clear the main layout
        self.graphLayout.clear()

        # Clear the additional ViewBoxes used for telluric and sky background
        try:
            self.telluric_vb.clear()
            self.sky_bkg_vb.clear()
        except AttributeError:
            pass

        self.data = {}
        self.img_colorbar = None
        self.wvlg_corrections = {
            "to_air": False,
            "to_vacuum": False,
            "heliocentric": False,
            "barycentric": False,
        }
        self.specsysModel.clear()

    def select_line_lists(self):
        SelectLineListsDialog(self)

    def velocity_plot(self):
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        indexes = self.specsysView.selectedIndexes()
        if indexes:
            # Indexes is a list of a single item in single-select mode.
            index = indexes[0]
            specsys = self.specsysModel.get_specsys(index)
            self.velocityWidget = VelocityPlot(
                self,
                z=specsys.redshift,
                lines=specsys.lines,
            )
            self.velocityWidget.show()
            # self.velocityWidget.plot_velocities()
        else:
            QtWidgets.QMessageBox.information(
                self,
                "No spectroscopic system selected",
                "Please select a spectroscopic system from the list "
                "to show the velocity plot.",
            )

    # Modify displayed data
    def smooth(self):
        """
        Get the number of pixels over which to smooth from the
        corresponding textbox. Then apply it the original 1D data
        loaded into memory.
        This is applied to the data in memory and not the displayed
        data, that way one can go back to lower values of smoothing.
        Otherwise, smoothing degrades information and one cannot
        "unsmooth" as that information is lost.
        """
        smoothing = int(self.textbox_for_smooth.text())
        self.statusBar.showMessage("Smoothing by {} pixels".format(smoothing), 2000)
        log.info("Smoothing {} pixels".format(smoothing))
        wvlg_sm, flux_sm, unc_sm = sf.smooth(
            self.data["wvlg_1D"].value,
            self.data["flux_1D"].value,
            unc=self.data["unc_1D"].value,
            smoothing=smoothing,
        )
        wvlg_sm = wvlg_sm * self.data["wvlg"].unit
        flux_sm = flux_sm * self.data["flux_1D"].unit
        unc_sm = unc_sm * self.data["unc_1D"].unit
        return wvlg_sm, flux_sm, unc_sm

    def apply_smoothing(self):
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        self.statusBar.showMessage("Smoothing...")
        try:
            wvlg_sm, flux_sm, unc_sm = self.smooth()
            # if self.mode == '1D':
            self.set_1D_displayed_data(wvlg_sm, flux_sm, unc_sm)
            self.calculate_1D_displayed_data_range()
            self.draw_1D_data()
            # TODO : implement 2D smoothing
            # elif self.mode == '2D':
            #     self.data['wvlg_2D_disp'] = x_sm
            #     self.data['flux_2D_disp'] = y_sm
            #     self.data['err_2D_disp'] = err_sm
            #     self.flux_2D_img.setImage(y_sm.T, levels=(self.data['q025'], self.data['q975']))
            #     self.err_2D_img.setImage(err_sm.T)
            #     # self.rect = QtCore.QRectF(x_sm[0], spat[0],
            #     # x_sm[-1]-x_sm[0], spat[-1]-spat[0])
            #     # self.flux_2D_img.setRect(self.rect)

        except ValueError:
            self.textbox_for_smooth.blockSignals(True)
            QtWidgets.QMessageBox.information(
                self,
                "Invalid smooth value",
                "Smoothing value must be convertible to integer",
            )
            self.textbox_for_smooth.blockSignals(False)
            self.textbox_for_smooth.setFocus()

    def reset_smoothing(self):
        """
        Display the original data the was read from the file or
        extracted from the 2D.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        self.set_1D_displayed_data(
            wvlg=self.data["wvlg_1D"],
            flux=self.data["flux_1D"],
            unc=self.data["unc_1D"],
        )
        self.calculate_1D_displayed_data_range()
        self.draw_1D_data()
        self.textbox_for_smooth.setText("1")
        # TODO : implement 2D smoothing

    def wvlg_to_air(self):
        """
        Convert the wavelength, assumed to be in vacuum, to air.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        if self.wvlg_corrections["to_air"]:
            QtWidgets.QMessageBox.information(
                self,
                "Invalid action",
                "You have already converted wavelength from vacuum " "to air.",
            )
        else:
            wvlg = self.data[f"wvlg_{self.mode}_disp"]
            wvlg_in_air = sf.vac_to_air(wvlg)
            self.data[f"wvlg_{self.mode}_disp"] = wvlg_in_air
            if not self.wvlg_corrections["to_vacuum"]:
                self.wvlg_corrections["to_air"] = True
            self.wvlg_corrections["to_vacuum"] = False
            self.draw_data()
            log.info("Converted wavelength from vacuum to air.")

    def wvlg_to_vacuum(self):
        """
        Convert the wavelength, assumed to be in air, to vacuum.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        if self.wvlg_corrections["to_vacuum"]:
            QtWidgets.QMessageBox.information(
                self,
                "Invalid action",
                "You have already converted wavelength from air " "to vacuum.",
            )
        else:
            wvlg = self.data[f"wvlg_{self.mode}_disp"]
            wvlg_in_vac = sf.air_to_vac(wvlg)
            self.data[f"wvlg_{self.mode}_disp"] = wvlg_in_vac
            if not self.wvlg_corrections["to_air"]:
                self.wvlg_corrections["to_vacuum"] = True
            self.wvlg_corrections["to_air"] = False
            self.draw_data()
            log.info("Converted wavelength from air to vacuum.")

    def wvlg_bary_correction(self):
        self.wvlg_radial_correction(kind="barycentric")

    def wvlg_helio_correction(self):
        self.wvlg_radial_correction(kind="heliocentric")

    def wvlg_radial_correction(self, kind):
        """
        Correct wavelength for barycentric or heliocentric motion.
        This uses information found in the header about the time of
        observation and the telescope location.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        if (
            self.wvlg_corrections["barycentric"]
            or self.wvlg_corrections["heliocentric"]
        ):
            QtWidgets.QMessageBox.information(
                self,
                "Invalid action",
                "You have already corrected for barycentric or " "heliocentric motion.",
            )
        elif self.data["header"] is None:
            QtWidgets.QMessageBox.information(
                self,
                "Invalid action",
                "You can only correct fits files for barycentric or "
                "heliocentric motion.",
            )
        else:
            try:
                wvlg = self.data[f"wvlg_{self.mode}_disp"]
                c = cst.c.to("km/s")
                vcorr = sf.calc_vel_corr(header=self.data["header"], kind=kind)
                wvlg_corr = wvlg * (1.0 + vcorr / c)
                self.data[f"wvlg_{self.mode}_disp"] = wvlg_corr
                self.wvlg_corrections["barycentric"] = True
                self.wvlg_corrections["heliocentric"] = True
                self.draw_data()
                log.info("Converted wavelength from {:s} motion".format(kind))
            except KeyError as e:
                log.error(e)
                QtWidgets.QMessageBox.information(
                    self,
                    "Invalid header",
                    f"Could not correct for {kind} motion :\n" "" + str(e),
                )

    # Ratios
    def calculate_ratio(self):
        """Calculates the ratio between the two selected wavelength"""
        try:
            nom = float((self.textbox_for_wvlg2.text()))
            denom = float((self.textbox_for_wvlg1.text()))
            ratio_value = float(nom / denom)
            if ratio_value < 1:
                ratio_value = 1.0 / ratio_value
            if ratio_value < 0:
                raise ValueError("Your ratio should never be negative")
            self.textbox_for_ratio.setText("{:.5f}".format(ratio_value))
        except ValueError:
            self.textbox_for_ratio.setText("Invalid input")

    def find_ratio_names(self):
        """Finds a list of close ratios given an error margin"""
        try:
            ratio_error_margin = float(self.textbox_for_ratio_error_margin.text())
            ratio_value = float(self.textbox_for_ratio.text())

            cond = abs(self.line_ratios["ratio"] - ratio_value) <= ratio_error_margin
            self.ratio_name_list.clear()
            for name in self.line_ratios[cond]["name"]:
                self.ratio_name_list.addItem(name)

        except ValueError:
            self.ratio_name_list.clear()
            self.ratio_name_list.addItem("Invalid input")

    # Systems
    def add_specsys_from_ratio(self):
        """
        Add a spectroscopic system from a given selected line ratio.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        try:
            chosen_ratio = self.ratio_name_list.selectedItems()[0]
        except IndexError:
            log.debug("Empty ratio list")
            return

        log.debug(f"Chosen ratio is: {chosen_ratio.text()}")
        if chosen_ratio:
            # Choose the first line name with [0]
            l_name = chosen_ratio.text().split("/")[0].strip()
            log.debug(f"Line name to search for is: {l_name}")

            cond = self.check_line_name(l_name)
            l_wvlg_rest = Quantity(self.abs_lines[cond]["wave"])[0]
            log.debug(f"Found corresponding wavelength: {l_wvlg_rest}")
            try:
                l1 = float(self.textbox_for_wvlg1.text())
                l2 = float(self.textbox_for_wvlg2.text())
                l_wvlg_obs = np.max([l1, l2])
                # Use the max wavelength because we chose the first
                # of the line names
                l_wvlg_obs = l_wvlg_obs * self.data["wvlg"].unit
            except Exception:
                QtWidgets.QMessageBox.information(
                    self,
                    "Invalid spectral system",
                    "Lambda 1 and Lambda 2 must be convertible to Quantity or float",
                )
                self.textbox_for_wvlg1.setFocus()
                return
            log.debug(f"Collecting observed wavelength: {l_wvlg_obs}")
            l_wvlg_rest = l_wvlg_rest.to(l_wvlg_obs.unit)
            z = l_wvlg_obs / l_wvlg_rest - 1.0
            log.debug(f"Calculated corresponding redshift: {z}")
            self.add_specsys(z=z, sys_type="abs")

    def add_specsys_from_line(self):
        """
        Add a spectroscopic system from a specific chosen line name.
        """
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return

        chosen_line = str(self.textbox_for_line.text())

        if not chosen_line:
            log.debug("Empty line textbox.")
            return

        else:
            log.debug(f"Chosen line is: {chosen_line}")
            cond = self.check_line_name(chosen_line)
            if cond is not None:
                l_wvlg_rest = Quantity(self.abs_lines[cond]["wave"])[0]
                log.debug(f"Found corresponding wavelength: {l_wvlg_rest}")
                try:
                    l_wvlg_obs = float(self.textbox_for_wvlg1.text())
                    l_wvlg_obs = l_wvlg_obs * self.data["wvlg"].unit
                except Exception:
                    QtWidgets.QMessageBox.information(
                        self,
                        "Invalid spectral system",
                        "Lambda 1 must be convertible to Quantity or float",
                    )
                    self.textbox_for_wvlg1.setFocus()
                    return
                log.debug(f"Collecting observed wavelength from Lambda 1: {l_wvlg_obs}")
                l_wvlg_rest = l_wvlg_rest.to(l_wvlg_obs.unit)
                z = (l_wvlg_obs / l_wvlg_rest).value - 1.0
                log.debug(f"Calculated corresponding redshift: {z}")
                self.add_specsys(z=z, sys_type="abs")

    def add_absorber(self):
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return
        try:
            z = self.textbox_for_z.text()
            log.debug("z is %s, reading from textbox for z", z)
            z = float(z)
            self.add_specsys(z, sys_type="abs")
        except ValueError:
            QtWidgets.QMessageBox.information(
                self,
                "Invalid spectral system",
                "Can't add system: z must be convertible to float",
            )

    def add_emitter(self):
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return
        try:
            z = self.textbox_for_z.text()
            log.debug("z is %s, reading from textbox for z", z)
            z = float(z)
            self.add_specsys(z, sys_type="em")
        except ValueError:
            QtWidgets.QMessageBox.information(
                self,
                "Invalid spectral system",
                "Can't add system: z must be convertible to float",
            )

    def add_specsys(self, z, sys_type="abs"):
        """
        Add a spectroscopic system at a given redshift and draw it
        on the 1D plot.
        """
        try:
            if sys_type == "abs":
                sys_type_str = "absorber"
                lines = join(self.abs_lines, self.fs_lines, join_type="outer")
            elif sys_type == "em":
                lines = self.em_lines
                sys_type_str = "emitter"
            color = self.get_color()
            specsys = SpecSystem(
                z=z,
                sys_type=sys_type,
                PlotItem=self.ax1D,
                color=color,
                lines=lines,
                show_fs=True,
            )
            self.statusBar.showMessage("Adding system at redshift %.5lf" % z, 2000)
            specsys.draw(
                xmin=self.data["wvlg_min"],
                xmax=self.data["wvlg_max"],
            )
            # Store the spectroscopic system in the model
            self.specsysModel.specsystems.append((True, specsys))
            self.specsysModel.sort()
            self.specsysModel.layoutChanged.emit()
            # Connect the custom signal 'edited' of the spectroscipic system
            # in order to update the model as soon as it receives the signal
            specsys.edited.connect(self.specsysModel.layoutChanged.emit)
            self.textbox_for_z.setText("")
            self.clear_color_from_available_list(color)
            log.info("Added %s at redshift %.5lf", sys_type_str, z)
        except ValueError as e:
            log.error(f"Can't add system: z must be convertible to float. Error : {e}")
            self.textbox_for_z.setFocus()

    def delete_specsys(self):
        if not self.data:
            log.debug("You pushed a button but did not load any data. Ignoring.")
            return
        indexes = self.specsysView.selectedIndexes()
        if indexes:
            # Indexes is a list of a single item in single-select mode.
            index = indexes[0]
            # Re-add the color to the available pool
            self.available_colors.append(self.specsysModel.get_color(index))
            self.specsysModel.delete(index)
            self.specsysModel.sort()
            # Clear the selection (as it is no longer valid).
            self.specsysView.clearSelection()

    def check_line_name(self, l_name):
        """
        Make sure line name exists in list of lines provided.
        """
        cond = [i for i, name in enumerate(self.abs_lines["name"]) if l_name in name]

        if len(self.abs_lines[cond]) == 0:
            QtWidgets.QMessageBox.information(
                self,
                "No lines found",
                "Could not find any line names associated with "
                f"{l_name}. Check line list provided.",
            )
            self.textbox_for_line.setFocus()
            return None
        elif len(self.abs_lines[cond]) >= 2:
            QtWidgets.QMessageBox.information(
                self,
                "Too many lines found",
                f"Found more than one line for {l_name}. " "Check line list provided.",
            )
            self.textbox_for_line.setFocus()
            return None
        else:
            log.debug(f"Found line: {self.abs_lines[cond]['name'][0]}")
            return cond


def main():
    app = QtWidgets.QApplication(sys.argv)
    main = MainWindow()
    main.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



import pandas as pd
import numpy as np
from astropy.units.quantity import Quantity
import astropy.units as u
import pyqtgraph as pg
from pathlib import Path
import logging

log = logging.getLogger(__name__)


def check_flux_scale(flux, unc):
    """
    Scale flux and unc to have them in reasonable units
    and allow y crosshair to work (otherwise the code considers
    it to be zero) and to allow histograms to be nicer
    """
    exponent = int(f"{np.std(flux).value:e}".split("e")[-1])
    if exponent >= 0 and exponent <= 2:
        # log.debug("Flux units are reasonable, no need to rescale to make them usable")
        pass
    else:
        log.debug(
            f"Flux and uncertainty need to be rescaled (exponent: {exponent}), rescaling them."
        )
        flux_rescaled = to_usable_units(flux)
        unc_rescaled = unc.to(flux_rescaled.unit)
        flux = flux_rescaled
        unc = unc_rescaled

    return flux, unc


def to_usable_units(input_array):
    """Return a copy of the input array in "usable" units by using
    the median of the array.
    For example, if an array's values is around 3.1 * 10^(-9) Jy
    it will return the array, normalized by 10^(-9).
    This function is relevant if the orders of magnitude spanned by the
    array is rather small (less than 2-4 orders of magnitude).
    It is used to avoid manipulating very small or very large numbers.

    Parameters
    ----------
    input_array : Quantity
        Array of Quantity to make "usable"

    Returns
    -------
    output_array
        Normalized array to "usable" values (order of magnitude ~1).
    """
    quant = np.std(input_array).value
    exponent = f"{quant:e}".split("e")[-1]
    new_unit = u.Unit(float(f"1e{int(exponent):d}") * input_array.unit)
    return input_array.to(new_unit)


def _quantity_to_at_least_1D_array(x):
    if isinstance(x, Quantity):
        x = np.atleast_1d(x.value)
    else:
        x = np.atleast_1d(x)
    return x


def set_up_linked_vb(pi):
    """
    Take a PlotItem and create a new ViewBox by the x axis
    """
    new_vb = pg.ViewBox()
    pi.scene().addItem(new_vb)
    new_vb.setXLink(pi)

    # Handle view resizing
    def updateViews():
        # view has resized; update auxiliary views to match
        new_vb.setGeometry(pi.vb.sceneBoundingRect())

        # need to re-update linked axes since this was called
        # incorrectly while views had different shapes.
        # (probably this should be handled in ViewBox.resizeEvent)
        new_vb.linkedViewChanged(pi.vb, new_vb.XAxis)

    updateViews()
    pi.vb.sigResized.connect(updateViews)
    return new_vb


def create_line_ratios(input_fname, sep=",", output_fname="line_ratio.csv", save=True):
    """
    Takes a line list and calculate all possible ratios between the
    line wavelengths. Then keep only 1 < ratio <= 2 and write them
    to a file.
    Input file must contain 2 columns for the name and the wavelength
    """
    lines = pd.read_csv(input_fname, sep=sep, comment="#")
    ratio = []
    ratio_name = []

    for n1, w1 in zip(lines["name"], lines["wvlg"]):
        for n2, w2 in zip(lines["name"], lines["wvlg"]):
            ratio_name.append("/".join([n1.strip(), n2.strip()]))
            ratio.append(w1 / w2)

    df_ratios = pd.DataFrame({"ratio": ratio, "name": ratio_name})
    usable_ratios = (df_ratios["ratio"] > 1) & (df_ratios["ratio"] <= 2)
    df_ratios = df_ratios[usable_ratios]
    df_ratios = df_ratios.sort_values("ratio")

    line_dir = Path(str(input_fname)).resolve().parent
    output_fname = line_dir / output_fname
    if save:
        df_ratios.to_csv(output_fname, index=False)
        log.info("Saved line ratios in {}".format(output_fname))

    return df_ratios



import zhunter.io as io
import numpy as np
import astropy.constants as cst
import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord, EarthLocation
from astropy.io import fits
from spectres import spectres
import logging

log = logging.getLogger(__name__)


def smooth(wvlg, flux, unc=None, smoothing=3):
    """
    A function to smooth a spectrum.
    """
    if smoothing <= 0:
        raise ValueError("Smoothing must be strictly positive")

    wvlg_regrid = np.linspace(wvlg.min(), wvlg.max(), int(len(wvlg) / smoothing))
    output = spectres(
        new_wavs=wvlg_regrid,
        spec_wavs=wvlg,
        spec_fluxes=flux,
        spec_errs=unc,
        fill=0,
        verbose=False,
    )
    if unc is None:
        flux_sm = output
        unc_sm = None
    else:
        flux_sm = output[0]
        unc_sm = output[1]

    return wvlg_regrid, flux_sm, unc_sm


def correct_lambda_for_radial_velocity(fname, kind="barycentric", mode="1D"):
    """
    Correct lambda for heliocentric or barycentric radial velocity shift.
    Return the corrected wavelength
    """
    c = cst.c.to("km/s")
    if mode == "1D":
        wvlg, _, _ = io.read_fits_1D_spectrum(fname)
    elif mode == "2D":
        wvlg, _, _, _ = io.read_fits_2D_spectrum(fname)
    else:
        raise ValueError
    vcorr = calc_vel_corr(header=fits.getheader(fname), kind=kind)
    wvlg_corr = wvlg * (1.0 + vcorr / c)
    return wvlg_corr


def calc_vel_corr(header, kind="barycentric"):
    """
    Calculates the radial velocity correction given an observing date, a telescope position
    and an object's RA and DEC in the sky (along with the reference frame).
    Returns the velocity correction for barycentric or heliocentric motion in km/s
    """
    ra2000 = header["RA"] * u.deg
    dec2000 = header["DEC"] * u.deg
    tel_lat = header["HIERARCH ESO TEL GEOLAT"] * u.deg
    tel_long = header["HIERARCH ESO TEL GEOLON"] * u.deg
    tel_alt = header["HIERARCH ESO TEL GEOELEV"] * u.m
    frame = header["RADECSYS"].lower()
    mjd = header["MJD-OBS"]
    exptime = header["EXPTIME"]

    coord = SkyCoord(ra2000, dec2000, frame=frame)
    date_obs = Time(
        mjd + exptime / (2.0 * 86400.0), format="mjd"
    )  # midpoint of observation
    tel_pos = EarthLocation.from_geodetic(lat=tel_lat, lon=tel_long, height=tel_alt)
    vel_corr = coord.radial_velocity_correction(
        kind=kind, obstime=date_obs, location=tel_pos
    )
    vel_corr = vel_corr.to("km/s")
    log.debug(
        "Velocity correction calculated: {:.3f} {:s}".format(
            vel_corr.value, vel_corr.unit
        )
    )

    return vel_corr


def air_to_vac(wavelength):
    """
    Implements the air to vacuum wavelength conversion described in eqn 65 of
    Griesen 2006
    """
    wlum = wavelength.to(u.um).value
    return (
        1 + 1e-6 * (287.6155 + 1.62887 / wlum**2 + 0.01360 / wlum**4)
    ) * wavelength


def vac_to_air(wavelength):
    """
    Griesen 2006 reports that the error in naively inverting Eqn 65 is less
    than 10^-9 and therefore acceptable.  This is therefore eqn 67
    """
    wlum = wavelength.to(u.um).value
    nl = 1 + 1e-6 * (287.6155 + 1.62887 / wlum**2 + 0.01360 / wlum**4)
    return wavelength / nl



from pathlib import Path
import logging

__version__ = "0.10.1"

log = logging.getLogger(__name__)

ROOT_DIR = Path(__file__).parents[0]
DIRS = {"ROOT": ROOT_DIR, "UI": ROOT_DIR / "ui", "DATA": ROOT_DIR / "data"}
log.debug("DIRS: {}".format(DIRS))



import pyqtgraph as pg
from PyQt6 import QtGui
import logging
from zhunter import DIRS

log = logging.getLogger(__name__)


def get_gradient(color, reverse=False):
    """Summary

    Parameters
    ----------
    color : QColor
        Description

    Returns
    -------
    TYPE
        Description
    """
    lighter_color = QtGui.QColor(color.name())
    lighter_color.setAlpha(64)  # out of 255
    grad = QtGui.QLinearGradient(0, 0, 0, 1)
    if reverse:
        grad.setColorAt(0.0, pg.mkColor(lighter_color))
        grad.setColorAt(1.0, pg.mkColor(color))
    else:
        grad.setColorAt(1.0, pg.mkColor(lighter_color))
        grad.setColorAt(0.0, pg.mkColor(color))
    return grad


def rgb_to_hex(red, green, blue):
    """Return color as #rrggbb for the given color values."""
    return "#%02x%02x%02x" % (red, green, blue)


def get_cblind_colors(rtype=dict, fmt="rgb"):
    """
    From color-blind friendly colors in Nature article
    https://www.nature.com/articles/nmeth.1618
    Returns a list or dictionnary of RGB values.
    """
    colors = {
        "Sky blue": (86, 180, 233),
        "Orange": (230, 159, 0),
        "Bluish green": (0, 158, 115),
        "Reddish purple": (204, 121, 167),
        "Vermillion": (213, 94, 0),
        "Blue": (0, 114, 178),
        "Yellow": (240, 228, 66),
    }

    if fmt == "hex":
        for n, RGB in colors.items():
            r, g, b = RGB
            colors[n] = rgb_to_hex(r, g, b)
    elif fmt == "rgb":
        # Convert to RGB fraction for python
        for n, RGB in colors.items():
            r, g, b = RGB
            colors[n] = (r / 255, g / 255, b / 255)
    else:
        raise ValueError("Unsupported format for colorblind colors.")

    if rtype == dict:
        return colors
    elif rtype == list:
        return list(colors.values())


KRAKEN = {
    "spec": "#EBEBEB",
    "unc": "#BC271B",
    "specsys": [
        "#4A9EBC",  # Lightblue
        "#ECCA54",  # Yellow
        "#C72A70",  # Pink
        "#C95D38",  # Orange
        "#92D754",  # Green
        "#8218BB",  # Purple
        # "#BC271B",  # Rust
        "#66CBA0",  # Teal
        "#B52EB0",  # Fuschia
        "#2D67EE",  # Blue
    ],
    "sky": "#696969",
    "crosshair": "#C6C6C6",
    "background": "#000000",
    "foreground": "#C6C6C6",
    "roi": "g",
    # "roi": "#3E8D26",  # green
}


CVD = {
    "spec": "white",
    "unc": "red",
    "specsys": get_cblind_colors(rtype=list, fmt="hex"),
    "sky": "gray",
    "crosshair": "lightgray",
    "background": "#000000",
    "foreground": "lightgray",
    "roi": "#0041FF",  # blue
}

OLD = {
    "spec": "white",
    "unc": "red",
    "specsys": [
        "#a6cee3",
        "#1f78b4",
        "#b2df8a",
        "#33a02c",
        "#fb9a99",
        "#e31a1c",
        "#fdbf6f",
        "#ff7f00",
        "#cab2d6",
        "#6a3d9a",
        "#ffff99",
        "#b15928",
    ],
    "sky": "gray",
    "crosshair": None,
    "background": "#000000",
    "foreground": "white",
    "roi": "g",
}

CYBERPUNK = {
    "spec": "white",
    "unc": "red",
    "specsys": ["#08F7FE", "#FE53BB", "#F5D300", "#41ff41", "#ff4141", "#9467bd"],
    "sky": "gray",
    "crosshair": "lightgray",
    "background": "#212946",
    "foreground": "lightgray",
    "roi": "g",
}

COLORS = {
    "kraken": KRAKEN,
    "cvd": CVD,
    "old": OLD,
    "cyberpunk": CYBERPUNK,
}



from astropy.io import fits
from astropy.io.ascii import read as ascii_read
from astropy.table import Table
import astropy.units as u
from astropy.nddata import StdDevUncertainty
from specutils import Spectrum1D
import numpy as np
import pandas as pd
import logging
from pathlib import Path

log = logging.getLogger(__name__)

ergscm2AA = u.def_unit(
    s="ergscm2AA",
    represents=u.Unit("erg s^-1 cm^-2 AA^-1"),
    format={"latex": r"\mathrm{erg\,s^{-1}\,cm^{-2}\,\mathring{A}^{-1}}"},
)
GREEK_LETTERS = ["alpha", "beta", "gamma", "delta"]
ERROR_KEYS = ["ERR", "NOISE", "SIGMA", "UNC"]
WAVE_KEYS = ["WAVE", "AWAV", "WVLG", "LAM"]
FLUX_KEYS = ["FLUX"]


def read_1D_data(fname):
    """
    A wrapper function to handle case for fits extension or generic txt
    or dat file
    """
    # Make sure fname is a Path instance
    if not isinstance(fname, Path):
        fname = Path(fname)

    if not fname.exists():
        raise Exception("File does not exist")

    fname_extension = fname.suffix
    # If file is compressed, check the original extension
    if fname.suffix in [".gz"]:
        fname_extension = Path(fname.stem).suffix
        if not fname_extension:
            raise IOError("No file extension found after stripping '.gz'")

    # Load data
    if fname_extension in [".fits", ".fit"]:
        spectrum = read_fits_1D_spectrum(fname)

    else:
        spectrum = read_generic_1D_spectrum(fname)

    return spectrum


def read_generic_1D_spectrum(fname, wave_unit=None, flux_unit=None):
    """
    Can read ECSV format.
    Ignores line starting with '#'
    Expects the following format:
        # this is a comment in the file which will be ignored
        wave, flux, error/uncertainty
        1, 10, 0.1
        2, 10, 0.1
        3, 10, 0.1
        4, 10, 0.1

    The error/uncertainty column is optional.
    The name of the columns should be the first uncommented line, with
    the same separator as the columns.
    You can specify the units of the column by directly following the
    column name with the unit in parenthesis:

        wave(AA), flux(erg/s/cm2/AA)

    Args:
        fname (TYPE): Description
        wave_unit (None, optional): Description
        flux_unit (None, optional): Description

    Returns:
        TYPE: Description
    """

    log.debug(f"Attempting to read file: {fname}")

    if not isinstance(fname, Path):
        fname = Path(fname)

    waveobs = None
    flux = None
    uncertainty = None

    if fname.suffix == ".ecsv":
        tab = ascii_read(fname)
        pandas = False
    else:
        pandas = True
        tab = pd.read_csv(
            fname,
            comment="#",
            dtype=float,
        )

    column_names = list(tab.columns)

    # Wavelength
    wave_key = __find_column_name(
        column_names,
        possible_names=WAVE_KEYS,
    )
    # Flux
    flux_key = __find_column_name(
        column_names,
        possible_names=FLUX_KEYS,
    )
    # Error
    uncertainty_key = __find_column_name(
        column_names,
        possible_names=ERROR_KEYS,
    )

    # If no column names, assume wave, flux, error/uncertainty
    if any(k is None for k in [wave_key, flux_key]):
        log.info(
            f"Could not find any reasonable column names for file {fname}."
            " Trying to extract data assuming default format wave, flux, uncertainty."
        )

        wave_key, flux_key, uncertainty_key = ("wave", "flux", "uncertainty")

        tab = pd.read_csv(
            fname,
            names=[wave_key, flux_key, uncertainty_key],
            comment="#",
            dtype=float,
        )

    # Try to get units from the column name
    if wave_key is not None:
        wave_unit = __parse_units_from_column_name(wave_key)
    if flux_key is not None:
        flux_unit = __parse_units_from_column_name(flux_key)

    # Have to turn to numpy because using pandas series
    if wave_key is not None:
        if pandas:
            waveobs = tab[wave_key].to_numpy()
        else:
            waveobs = tab[wave_key]
            wave_unit = waveobs.unit
    if flux_key is not None:
        if pandas:
            flux = tab[flux_key].to_numpy()
        else:
            flux = tab[flux_key]
            flux_unit = flux.unit
    if uncertainty_key is not None:
        if pandas:
            uncertainty = tab[uncertainty_key].to_numpy()
            # If the column didn't exist, pandas fills with NaNs by default
            if all(np.isnan(uncertainty)):
                uncertainty = None
        else:
            uncertainty = tab[uncertainty_key]

    # Default units if no units were specified
    if wave_unit is None:
        wave_unit = u.Unit("AA")
        log.info("No units specified for wave, assuming Angstrom.")
    if flux_unit is None:
        flux_unit = u.Unit("adu")
        log.info("No units specified for flux, assuming ADU.")

    if flux is not None and waveobs is not None:
        # Create the spectrum object
        spectrum = Spectrum1D(
            spectral_axis=waveobs * wave_unit,
            flux=flux * flux_unit,
        )
        if uncertainty is not None:
            spectrum.uncertainty = StdDevUncertainty(uncertainty)
        else:
            log.warning(f"No error/uncertainty found in file {fname}.")
    else:
        raise Exception("Unknown file format")
    return spectrum


def read_fits_2D_spectrum(fname, verbose=False):
    """
    A function to read data from a 2D spectrum.
    First looks for a HDU with a name containing 'FLUX'.
    If it doesn't find any, it searches through HDUs that are either
    ImageHDU or PrimaryHDU for data in the form of a 2D array.
    It also searches for an error/uncertainty HDU by matching a name in
    ["ERR", "NOISE", "SIGMA", "UNC"]
    If no error/uncertainty is founds, it returns 0.
    Finally, it builds the spatial and spectral dimensions by reading
    the header of the HDU where it found the flux.
    """

    log.debug(f"Attempting to read file: {fname}")

    with fits.open(fname) as hdulist:
        hdu_names = [hdu.name for hdu in hdulist]

        flux_hdu_name = __find_column_name(
            hdu_names,
            possible_names=FLUX_KEYS,
        )

        if flux_hdu_name:
            hdu = hdulist[flux_hdu_name]
            flux = hdu.data
        else:
            log.info(
                f"Couldn't find an HDU with a name containing any of {FLUX_KEYS}. "
                "Looking for primary or image HDUs containing 2D arrays..."
            )
            for hdu in hdulist:
                if isinstance(
                    hdu, (fits.hdu.image.PrimaryHDU, fits.hdu.image.ImageHDU)
                ):
                    flux = hdu.data
                    if flux is None:
                        continue
                    elif len(flux.shape) == 2:
                        log.debug(f"Found HDU: {hdu.name} which contains a 2D array")
                        break

        if flux is None or len(flux.shape) != 2:
            raise Exception("Could not find a 2D flux array in this fits file.")

        # Error/uncertainty
        uncertainty_hdu_name = __find_column_name(
            hdu_names,
            possible_names=ERROR_KEYS,
        )

        if uncertainty_hdu_name:
            uncertainty = hdulist[uncertainty_hdu_name].data
        else:
            log.info(
                f"Couldn't find an HDU with a name containing any of {ERROR_KEYS}."
                " Setting errors/uncertainties to 0."
            )
            uncertainty = np.zeros(flux.shape)

    flux_unit = hdu.header.get("BUNIT")
    if flux_unit is not None:
        flux_unit = u.Unit(flux_unit)
    else:
        flux_unit = u.Unit()
    flux = flux * flux_unit
    uncertainty = uncertainty * flux_unit

    spixels = np.arange(flux.shape[0])
    wpixels = np.arange(flux.shape[1])

    # Spatial dimension
    spatial_unit = __get_units(header=hdu.header, axis=2, default_units=u.arcsec)
    spatial_constructor = __get_constructor(header=hdu.header, axis=2)
    spatial = spatial_constructor(spixels)
    if spatial_unit is not None:
        spatial = spatial * spatial_unit
    else:
        spatial = spatial * u.Unit()

    # Spectral dimension
    wave_unit = __get_wavelength_units(header=hdu.header)
    wave_constructor = __get_wavelength_constructor(header=hdu.header)
    waveobs = wave_constructor(wpixels)
    if wave_unit is not None:
        waveobs = waveobs * wave_unit
    else:
        waveobs = waveobs * u.Unit()

    return waveobs, spatial, flux, uncertainty


def read_fits_1D_spectrum(fname):
    """
    Adapted from iSpec:
        https://github.com/marblestation/iSpec/blob/master/ispec/spectrum.py
    Reads the 'PRIMARY' HDU of the FITS file, considering that it contains the fluxes.
    The wavelength are derived from the headers, if not possible it checks if the
    data from the HDU contains 2 axes and takes the first as the wavelength
    and the second as the flux.
    It tries to find the errors/uncertainties in other HDU by checking the names and the length,
    Finally, if nothing has worked, it searches for a binary table.
    It looks for a wavelength column with a name containing: 'WAVE', 'AWAV', 'WVLG', 'LAM'
    a flux column with a name containing: 'FLUX'
    and an error/uncertainty column with a name containing: 'ERR', 'NOISE', 'SIGMA', 'UNC'

    returns a Spectrum1D object
    """

    log.debug(f"Attempting to read file: {fname}")

    with fits.open(fname) as hdulist:
        # By default start with PRIMARY HDU
        data = hdulist["PRIMARY"].data
        hdr = hdulist["PRIMARY"].header

        waveobs = None
        wave_unit = None
        flux = None
        flux_unit = None
        uncertainty = None

        if data is not None and (
            isinstance(
                hdulist["PRIMARY"], (fits.hdu.image.PrimaryHDU, fits.hdu.image.ImageHDU)
            )
        ):
            # If data has more than one dimension and
            # each dimension has more than one length (not just a single value)
            # This is to detect 2D spectra where one axis is wavelength
            # and the other is spatial dimension (arcsec)
            if hdr["NAXIS"] > 1 and all(dim > 1 for dim in data.shape):
                raise Exception("Can only read 1D fits spectra")

            flux = data.flatten()
            pixels = np.arange(len(flux))

            flux_unit = __get_flux_units(
                header=hdr,
                axis=2,
                default_units=ergscm2AA,
            )
            wave_unit = __get_wavelength_units(header=hdr)
            wave_constructor = __get_wavelength_constructor(header=hdr)
            waveobs = wave_constructor(pixels)

            if waveobs is None:
                # If could not build a constructor from WCS in header
                # try to see if data array has more than one dimension
                if len(data.shape) > 1:
                    log.info(
                        "Assuming 2D input with 1st dimension representing the spectral axis."
                    )
                    # No valid WCS, try assuming first axis is the wavelength axis
                    if hdr.get("CUNIT1") is not None:
                        waveobs = data[0, :]
                        flux = data[1, :]
                        if data.shape[0] > 2:
                            uncertainty = data[2, :]
                    else:
                        raise Exception("Unknown FITS file format")
                else:
                    raise Exception("Unknown FITS file format")

            # Try to find the errors/uncertainties in the extensions (HDU different than the PRIMARY):
            for hdu in hdulist:
                name = hdu.name.upper()
                if (
                    "PRIMARY" in name
                    or len(hdu.data.flatten()) != len(flux)
                    or isinstance(hdu, fits.hdu.table.BinTableHDU)
                ):
                    continue
                if name in ["IVAR", "IVARIANCE"]:
                    uncertainty = np.sqrt(1.0 / hdu.data.flatten())  # Not sure
                    # uncertainty = 1. / hdulist[i].data.flatten()
                    break
                elif name in ["VAR", "VARIANCE"]:
                    uncertainty = np.sqrt(hdu.data.flatten())  # Not sure
                    # uncertainty = hdulist[i].data.flatten()
                    break
                # Write it this way in case some people use ERR and others ERROR or ERRS
                elif name in ["NOISE", "SIGMA"] or "ERR" in name:
                    uncertainty = hdu.data.flatten()
                    break

        # If not data in Primary HDU
        elif data is None:
            log.info("Data is not in PRIMARY HDU, searching for binary tables...")
            # Try to find a binary table with the right columns
            # Stop after having found the first table that works
            for hdu in hdulist:
                if isinstance(hdu, fits.hdu.table.BinTableHDU):
                    bin_tab = Table.read(hdu)
                    column_names = list(bin_tab.columns)

                    # Wavelength
                    wave_key = __find_column_name(
                        column_names,
                        possible_names=WAVE_KEYS,
                    )
                    waveobs = bin_tab[wave_key]
                    if waveobs is not None:
                        # Need to call .flatten() here because sometimes
                        # we have an array inside another as a single element
                        # in data from eso archive
                        waveobs = np.array(waveobs).flatten()
                        wave_unit = bin_tab[wave_key].unit

                    # Flux
                    flux_key = __find_column_name(
                        column_names,
                        possible_names=FLUX_KEYS,
                    )
                    flux = bin_tab[flux_key]
                    if flux is not None:
                        # Need to call .flatten() here (see wavelength above)
                        flux = np.array(flux).flatten()
                        flux_unit = bin_tab[flux_key].unit

                    # Error
                    uncertainty_key = __find_column_name(
                        column_names,
                        possible_names=ERROR_KEYS,
                    )
                    uncertainty = bin_tab[uncertainty_key]
                    if uncertainty is not None:
                        # Need to call .flatten() here (see wavelength above)
                        uncertainty = np.array(uncertainty).flatten()
                        # Don't check for error/uncertainty units, Spectrum1D assumes they
                        # have same units as flux

    # Default units if no units were found
    if wave_unit is None:
        wave_unit = u.AA
    if flux_unit is None:
        flux_unit = ergscm2AA

    if flux is not None and waveobs is not None:
        # Create the spectrum object
        spectrum = Spectrum1D(
            spectral_axis=waveobs * wave_unit,
            flux=flux * flux_unit,
        )
        if uncertainty is not None:
            spectrum.uncertainty = StdDevUncertainty(uncertainty)

        return spectrum
    else:
        # If didn't return a spectrum with Primary
        # Or didn't find any binary table with the right columns
        raise Exception("Unknown FITS file format")


def convert_to_ecsv(fname, delimiter=" ", units=None):
    """Convert an ascii file to ECSV format which stores meta data
    like units. You can specify the units of each column with the
    units argument. For more information about this recommended format
    see https://docs.astropy.org/en/stable/io/ascii/ecsv.html

    Args:
        fname (str or Path): Name of the file
        delimiter (str, optional): Can be ',' or ' '
        units (list, optional): List of units or strings representing
        valid units. For columns that don't have units, use ''.
    """

    # Make sure fname is a Path instance
    if not isinstance(fname, Path):
        fname = Path(fname)

    tab = ascii_read(fname)
    if units is not None:
        if len(units) != len(tab.columns):
            log.error("Please specify as many units as there are columns")
        else:
            for col, unit in zip(tab.columns, units):
                tab[col].unit = u.Unit(unit)
    log.info(f"Converting file {fname} to ecsv format")
    tab.write(fname.with_suffix(".ecsv"), delimiter=delimiter, overwrite=True)


def convert_line_name_to_latex(lname):
    if any(greek.lower() in lname.lower() for greek in GREEK_LETTERS):
        elem, wave = lname.split("_")
        latex_string = elem + rf"$\{wave}$"
        return latex_string
    else:
        n_fine_struct = lname.count("*")
        rad, wave = lname.strip("*").split("_")

        # Strip blank spaces
        rad = rad.strip()
        wave = wave.strip()

        elem = None
        for roman in ["I", "V"]:
            if roman in rad:
                # If element is Iodine (I) or Vanadium (V)
                if roman in rad[0]:
                    elem = rad[0]
                    ion = rad[1:]
                else:
                    elem = rad.split(roman)[0]
                    ion = rad.strip(elem)
                break

        if elem is None:
            log.error(f"Could not parse line name {lname} to convert to latex")
            return lname

        latex_string = (
            r"\hbox{"
            + elem
            + rf"\,\textsc{{{ion.lower()}}}"
            + n_fine_struct * "*"
            + rf"$\lambda{wave}$"
            + "}"
        )
        return latex_string


def __parse_units_from_column_name(col_name):
    """Look for units in the name of a column by searching for a format
    as: wave(nm)

    Args:
        col_name (str): column name to search

    Returns:
        astropy.Unit: The unit found or None.
    """
    try:
        units_str = col_name.split("(")[1].strip(")")
        units = u.Unit(units_str)
        log.debug(f"Parsed unit {units_str} from column name '{col_name}'")
    except Exception:
        units = None
        log.debug(f"No unit could be parsed from column name '{col_name}'")
    return units


def __get_units(header, axis=2, default_units=None):
    # Check for flux units
    try:
        cunit2 = header[f"CUNIT{axis}"].strip().lower()
        units = u.Unit(cunit2)
    except KeyError:
        if default_units is not None:
            log.info(f"No unit found in header, assuming {default_units}")
            units = default_units
        else:
            units = None
            log.info(
                "No unit found in header and no default units provided,"
                " returning None."
            )
    return units


def __get_flux_units(header, axis=2, default_units=None):
    # Check for flux units
    cunit2 = header.get(f"CUNIT{axis}")
    bunit = header.get("BUNIT")

    if cunit2:
        if cunit2 == "ADU":
            cunit2 = "adu"
        log.debug(f"Using CUNIT: {cunit2} for flux units")
        flux_units = u.Unit(cunit2)
    elif bunit:
        if bunit == "ADU":
            bunit = "adu"
        log.debug(f"Using BUNIT: {bunit} for flux units")
        flux_units = u.Unit(bunit)
    elif default_units:
        log.info(f"No unit found in header, assuming {default_units}")
        flux_units = default_units
    else:
        flux_units = None
        log.info(
            "No unit found in header and no default units provided," " returning None."
        )
    return flux_units


def __get_wavelength_units(header, waxis=1, default_units=u.AA):
    # Check for wavelength units
    try:
        cunit1 = header[f"CUNIT{waxis}"].strip().lower()
        # If the unit is angstroms, convert to angstrom otherwise
        # astropy won't understand
        if cunit1 == "angstroms":
            cunit1 = "angstrom"
        wave_unit = u.Unit(cunit1)
    except KeyError:
        log.info(
            f"No unit found in header for wavelength (axis {waxis}), assuming {default_units}"
        )
        wave_unit = default_units
    return wave_unit


def __get_constructor(header, axis=2):
    """
    Create a function by reading the header that can construct a
    physical axis if provided a pixel array.

    Args:
        header (astropy Header): Header from wich to read.
        axis (int, optional): The axis number in the header.
        By default this is 2.

    Returns:
        func: function that takes in a pixel array and outputs the
        corresponding physical array
    """
    # Try to read World Coordinate System (WCS) that defines the grid
    if header.get(f"CD{axis}_{axis}") is not None:
        step = header[f"CD{axis}_{axis}"]
        log.debug("Using the FITS CD matrix.")
    elif header.get(f"CDELT{axis}") is not None:
        step = header[f"CDELT{axis}"]
        log.debug("Using the FITS CDELT value.")
    else:
        step = None
        log.info(
            f"No CDELT or CD in header for axis {axis}."
            " Returning a constructor that yields None."
        )
        return lambda x: None

    if step:
        start = header[f"CRVAL{axis}"]
        reference_pixel = header[f"CRPIX{axis}"]
        log.debug(f"PIX={reference_pixel} VAL={start} DELT={step}")

    constructor = lambda x: ((x - reference_pixel + 1) * step + start)

    return constructor


def __get_wavelength_constructor(header, waxis=1):
    """
    Create a function by reading the header that can construct a
    wavelength axis if provided a pixel array.

    Args:
        header (astropy Header): Header from wich to read.
        waxis (int, optional): The axis of the wavelength in the header.
        By default this is 1.

    Returns:
        func: function that takes in a pixel array and outputs the
        corresponding wavelength
    """
    # Try to read World Coordinate System (WCS) that defines the wavelength grid
    if header.get(f"CD{waxis}_{waxis}") is not None:
        wave_step = header[f"CD{waxis}_{waxis}"]
        log.debug("Using the FITS CD matrix.")
    elif header.get(f"CDELT{waxis}") is not None:
        wave_step = header[f"CDELT{waxis}"]
        log.debug("Using the FITS CDELT value.")
    else:
        wave_step = None
        log.info(
            f"No CDELT or CD in header for axis {waxis}."
            " Returning a constructor that yields None."
        )
        return lambda x: None

    # if found information in header
    wave_base = header[f"CRVAL{waxis}"]
    reference_pixel = header[f"CRPIX{waxis}"]
    log.debug(f"PIX={reference_pixel} VAL={wave_base} DELT={wave_step}")

    # Deal with logarithmic wavelength binning if necessary
    if header.get("WFITTYPE") == "LOG-LINEAR":
        constructor = lambda x: 10 ** (
            (x - reference_pixel + 1) * wave_step + wave_base
        )
        log.info("Log scale for wavelength constructor.")
    else:
        constructor = lambda x: ((x - reference_pixel + 1) * wave_step + wave_base)

    return constructor


def __find_column_name(column_names, possible_names):
    """
    This function checks if any of the names in colnames are found
    in a list of possible keys.
    If no name matches the list of possible names, None is returned.

    Args:
        column_names (list): list of column names to search through.

    Returns:
        str or None: column name that matched
    """

    # Get a list of the names in column_names that match the possible names
    found_names = [
        k for k in column_names if any(pk in k.upper() for pk in possible_names)
    ]

    if len(found_names) == 1:
        # If only one column name found
        name = found_names[0]
    elif len(found_names) > 1:
        # If more than one column name matches, take the shortest name, this is arbitrary
        name = min(found_names, key=len)
        log.warning(
            f"More than one column name: {found_names} matched the possible names: {possible_names}."
            f" Using shortest name: {name}"
        )
    else:
        log.debug(f"Did not find any name matching {possible_names} in {column_names}")
        # If no column names matched, set the array to None
        name = None

    return name



from PyQt6 import uic
from PyQt6 import QtWidgets
from zhunter import DIRS


class KeyBindingHelpDialog(QtWidgets.QDialog):
    def __init__(self, parent):
        super(KeyBindingHelpDialog, self).__init__(parent)
        self.parent = parent
        uic.loadUi(DIRS["UI"] / "key_bindings.ui", self)
